import logging
from typing import Optional

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output, OWBaseWidget

from tomwer.gui.control.reducedarkflatselector import ReduceDarkFlatSelectorDialog

logger = logging.getLogger(__name__)


class ReduceDarkFlatSelectorOW(OWBaseWidget, openclass=True):
    name = "reduce dark-flat selector"
    id = "orangecontrib.widgets.tomwer.control.ReduceDarkFlatSelectorOW.ReduceDarkFlatSelectorOW"
    description = "Allow user to select one or several reduced dark / flat"
    icon = "icons/reduced_darkflat_selector.svg"
    priority = 242
    keywords = [
        "tomography",
        "selection",
        "tomwer",
        "scan",
        "data",
        "reduce",
        "dark",
        "flat",
        "reduced",
    ]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _configuration = Setting(tuple())

    class Inputs:
        in_reduced_frames = Input(
            name="reduced frames",
            type=dict,
            doc="dict of containing reduced frames (either dark or flat)",
            multiple=True,
        )

    class Outputs:
        out_reduced_darks = Output(
            name="reduced dark(s)",
            doc="dict of containing reduced darks(s)",
            type=dict,
        )
        out_reduced_flats = Output(
            name="reduced flat(s)",
            type=dict,
            doc="dict of containing reduced flat(s)",
        )

    def __init__(self, parent=None):
        """ """
        super().__init__(parent)

        self._dialog = ReduceDarkFlatSelectorDialog(parent=self)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.addWidget(self._dialog)

        self._dialog.setConfiguration(self._configuration)

        # connect signal / slot
        self._dialog.sigSelectActiveAsDarks.connect(self._sendDarks)
        self._dialog.sigSelectActiveAsFlats.connect(self._sendFlats)
        self._dialog.sigUpdated.connect(self._updateConfiguration)

    def _updateConfiguration(self):
        self._configuration = self.getConfiguration()

    def _sendDarks(self, darks: dict):
        self.Outputs.out_reduced_darks.send(darks)

    def _sendFlats(self, flats: dict):
        self.Outputs.out_reduced_flats.send(flats)

    @Inputs.in_reduced_frames
    def addReduceFrames(self, reduce_frames: Optional[dict], *args, **kwargs):
        if reduce_frames is not None:
            self._dialog.addReduceFrames(reduce_frames)

    # expose API
    def getConfiguration(self) -> tuple:
        return self._dialog.getConfiguration()

    def setConfiguration(self, configuration: tuple) -> None:
        if configuration is None:
            return
        return self._dialog.setConfiguration(configuration)



from typing import Optional
from silx.gui import qt
from silx.gui.colors import Colormap
from tomwer.gui import icons as tomwer_icons
from tomwer.core.utils.char import ALPHA_CHAR


class AlphaChannelWidget(qt.QWidget):
    """
    Widget to select alpha level of the stiched image and of the background
    """

    sigAlphaImgChanged = qt.Signal()
    """signal emit when the alpha of the stitched image changed"""
    sigAlphaBackgroundChanged = qt.Signal()
    """signal emit when the alpha of the background image changed"""

    def __init__(self, parent=None, *args, **kwargs) -> None:
        super().__init__(parent, *args, **kwargs)
        self.setLayout(qt.QFormLayout())
        self._alphaImgSlider = qt.QSlider(qt.Qt.Horizontal, self)
        self._alphaImgSlider.setMinimumWidth(40)
        self._alphaImgSlider.setRange(0, 100)
        self.layout().addRow(f"{ALPHA_CHAR} image", self._alphaImgSlider)
        self._alphaBackgroundSlider = qt.QSlider(qt.Qt.Horizontal, self)
        self._alphaBackgroundSlider.setMinimumWidth(40)
        self._alphaBackgroundSlider.setRange(0, 100)
        self.layout().addRow(f"{ALPHA_CHAR} overlay", self._alphaBackgroundSlider)

        # connect signal / slot
        self._alphaImgSlider.valueChanged.connect(self._alphaImgchanged)
        self._alphaBackgroundSlider.valueChanged.connect(self._alphaBackgroundchanged)

    def _alphaImgchanged(self, *args, **kwargs):
        self.sigAlphaImgChanged.emit()

    def _alphaBackgroundchanged(self, *args, **kwargs):
        self.sigAlphaBackgroundChanged.emit()

    def setAlphaStitchedImg(self, alpha: float):
        self._alphaImgSlider.setValue(int(alpha * 100.0))

    def getAlphaStitchedImg(self) -> float:
        return self._alphaImgSlider.value() / 100.0

    def setAlphaBackgroundImg(self, alpha: float):
        self._alphaBackgroundSlider.setValue(int(alpha * 100.0))

    def getAlphaBackgroundImg(self) -> float:
        return self._alphaBackgroundSlider.value() / 100.0


class StitchAndBackgroundAlphaMixIn:
    """
    Mix in class for windows displaying a stitched image with or without the background
    """

    LEGEND_STITCHED_FRAME = "stitched frame"
    LEGEND_BACKGROUND = "background"

    def __init__(self) -> None:
        self._stitched_image = None
        self._composition_background = None
        # background action to show / hide the backgrounds
        self._backGroundAction = qt.QAction("overlay", self)
        self._backGroundAction.setToolTip("show / hide overlay on top of image")
        self._backGroundAction.setCheckable(True)
        save_icon = tomwer_icons.getQIcon("background")
        self._backGroundAction.setIcon(save_icon)
        self.__firstPlot = True
        self._backgroundColormap = Colormap(name="viridis")

        # build
        self._alphaChannelWidget = AlphaChannelWidget(parent=self)

        # connect signal / slot
        self._backGroundAction.toggled.connect(self._update)
        self._alphaChannelWidget.sigAlphaImgChanged.connect(self._updateAlphaImg)
        self._alphaChannelWidget.sigAlphaBackgroundChanged.connect(
            self._updateAlphaBackgroundImg
        )

    def getStitchedImgAlpha(self) -> float:
        """
        :return: alpha value to set to the stitched image. If only the image is displayed then will be 1.0 else will be the value defined by the "alpha channel' widget
        """
        if self._backGroundAction.isChecked():
            return self._alphaChannelWidget.getAlphaStitchedImg()
        else:
            return 1.0

    def getBackgroundImgAlpha(self) -> Optional[float]:
        """return 0.0 if there is no background image"""
        if self._backGroundAction.isChecked():
            return self._alphaChannelWidget.getAlphaBackgroundImg()
        else:
            return 0.0

    def _updateAlphaImg(self, *args, **kwargs):
        alpha_img = self.getImage(self.LEGEND_STITCHED_FRAME)
        if alpha_img is not None:
            alpha_img.setAlpha(self.getStitchedImgAlpha())

    def _updateAlphaBackgroundImg(self, *args, **kwargs):
        background_img = self.getImage(self.LEGEND_BACKGROUND)
        if background_img is not None:
            background_img.setAlpha(self.getBackgroundImgAlpha())

    def _update(self):
        """
        update display. Allow to handle modification of alpha values and to dsiplay or not the background / composition
        """
        if self._stitched_image is None:
            return
        if self._backGroundAction.isChecked():
            self.addImage(
                self._stitched_image,
                legend=self.LEGEND_STITCHED_FRAME,
                resetzoom=self.__firstPlot,
            )
            image_data = self.getImage(legend=self.LEGEND_STITCHED_FRAME)
            image_data.setAlpha(self.getStitchedImgAlpha())
            self.addImage(
                self._composition_background,
                legend=self.LEGEND_BACKGROUND,
                resetzoom=self.__firstPlot,
                colormap=self._backgroundColormap,
            )
            background_data = self.getImage(legend=self.LEGEND_BACKGROUND)
            background_data.setAlpha(self.getBackgroundImgAlpha())
        else:
            if self.getImage(self.LEGEND_BACKGROUND) is not None:
                self.removeImage(self.LEGEND_BACKGROUND)
            self.addImage(
                self._stitched_image,
                legend=self.LEGEND_STITCHED_FRAME,
                resetzoom=self.__firstPlot,
            )
            background_data = self.getImage(legend=self.LEGEND_STITCHED_FRAME)
            background_data.setAlpha(self.getStitchedImgAlpha())
        self.setActiveImage(legend=self.LEGEND_STITCHED_FRAME)
        self.__firstPlot = False

    # expose API
    def setAlphaBackgroundImg(self, value):
        self._alphaChannelWidget.setAlphaBackgroundImg(value)

    def setAlphaStitchedImg(self, value):
        self._alphaChannelWidget.setAlphaStitchedImg(value)



import logging

from nabu.stitching import config as stitching_config
from nabu.stitching.config import StitchingType
from nxtomomill.io.utils import convert_str_to_bool
from silx.gui import qt

from tomwer.core.scan.nxtomoscan import NXtomoScan, NXtomoScanIdentifier
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.gui.utils.inputwidget import OutputVolumeDefinition
from tomwer.gui.qlefilesystem import QLFileSystem

_logger = logging.getLogger(__name__)


class _PreProcessingOutput(qt.QWidget):
    """
    Define output settings for the pre processing z stitching
    """

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QFormLayout())
        # TODO: check if the widget with output .nx file exists somewhere
        self._outputFile = QLFileSystem("stitching/stitched.nx", self)
        self.layout().addRow("output nexus file", self._outputFile)
        self._outputDataPath = qt.QLineEdit("entry0000", self)
        self.layout().addRow("output data path", self._outputDataPath)

    def getUrl(self) -> str:
        return NXtomoScanIdentifier(
            object=NXtomoScan,
            hdf5_file=self._outputFile.text(),
            entry=self._outputDataPath.text(),
        )

    def setUrl(self, url: str):
        try:
            identifier = NXtomoScanIdentifier.from_str(url)
        except Exception as e:
            _logger.warning(f"Fail to create an identifier from {url}. Error is {e}")
        else:
            self._outputFile.setText(identifier.file_path)
            self._outputDataPath.setText(identifier.data_path)


class _PostProcessingOutput(OutputVolumeDefinition):
    """
    Define output settings for the post processing z stitching
    """

    def getUrl(self) -> str:
        return self.getOutputVolumeIdentifier().to_str()

    def setUrl(self, url: str):
        self.setOutputVolumeIdentifier(url)


class StitchingOutput(qt.QWidget):
    """
    Defines the output of the stitching
    """

    sigChanged = qt.Signal()

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.__stitching_type = None
        self.setLayout(qt.QVBoxLayout())
        self._preProcOutput = _PreProcessingOutput(parent=self)
        self.layout().addWidget(self._preProcOutput)
        self._postProcOutput = _PostProcessingOutput(parent=self)
        self.layout().addWidget(self._postProcOutput)
        self._overwritePB = qt.QCheckBox("overwrite", self)
        self._overwritePB.setChecked(True)
        self.layout().addWidget(self._overwritePB)

        # add a vertical spacer for display
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        self._updateOutputForStitchingType(StitchingType.Z_PREPROC)

    def _updateOutputForStitchingType(self, stitching_type):
        self.__stitching_type = StitchingType.from_value(stitching_type)
        self._preProcOutput.setVisible(self.__stitching_type is StitchingType.Z_PREPROC)
        self._postProcOutput.setVisible(
            self.__stitching_type is StitchingType.Z_POSTPROC
        )

    def getConfiguration(self) -> dict:
        config = {
            stitching_config.OUTPUT_SECTION: {
                stitching_config.OVERWRITE_RESULTS_FIELD: self._overwritePB.isChecked()
            },
        }
        if self.__stitching_type is StitchingType.Z_POSTPROC:
            config.update(
                {
                    stitching_config.Z_POST_PROC_SECTION: {
                        stitching_config.OUTPUT_VOLUME: self._postProcOutput.getUrl()
                    },
                }
            )
            return config

        elif self.__stitching_type is StitchingType.Z_PREPROC:
            url = self._preProcOutput.getUrl()
            try:
                scan = ScanFactory.create_tomo_object_from_identifier(url)
            except Exception:
                scan = None

            if not isinstance(scan, NXtomoScan):
                _logger.info("Failed to create an HDFTomoscan from url")
                return None
            else:
                config.update(
                    {
                        stitching_config.Z_PRE_PROC_SECTION: {
                            stitching_config.DATA_FILE_FIELD: scan.master_file,
                            stitching_config.DATA_PATH_FIELD: scan.entry,
                        },
                    }
                )
                return config
        else:
            raise NotImplementedError

    def setConfiguration(self, config: dict):
        stitching_type = config.get(stitching_config.STITCHING_SECTION, {}).get(
            stitching_config.STITCHING_TYPE_FIELD, None
        )
        if stitching_type is not None:
            self._updateOutputForStitchingType(stitching_type)

        output_volume = config.get(stitching_config.Z_POST_PROC_SECTION, {}).get(
            stitching_config.OUTPUT_VOLUME, None
        )
        if output_volume:
            self._postProcOutput.setUrl(output_volume)

        location = config.get(stitching_config.Z_PRE_PROC_SECTION, {}).get(
            stitching_config.DATA_FILE_FIELD, None
        )
        data_path = config.get(stitching_config.Z_PRE_PROC_SECTION, {}).get(
            stitching_config.DATA_PATH_FIELD, None
        )
        if location is not None:
            identifier = NXtomoScanIdentifier(
                object=NXtomoScan, hdf5_file=location, entry=data_path
            )
            self._preProcOutput.setUrl(identifier.to_str())

        overwrite = config.get(stitching_config.OUTPUT_SECTION, {}).get(
            stitching_config.OVERWRITE_RESULTS_FIELD, None
        )
        if overwrite is not None:
            overwrite = convert_str_to_bool(overwrite)
            self._overwritePB.setChecked(overwrite)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "22/01/2017"


import shutil
import tempfile

import pytest
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.utils.scanutils import MockEDF
from tomwer.gui.control.scanselectorwidget import ScanSelectorWidget
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestScanSelector(TestCaseQt):
    """
    Simple test for the ScanSelectorWidget
    """

    def setUp(self):
        self._folder1 = tempfile.mkdtemp()
        self._folder2 = tempfile.mkdtemp()
        self._folder3 = tempfile.mkdtemp()
        for _folder in (self._folder1, self._folder2, self._folder3):
            MockEDF.mockScan(scanID=_folder, nRadio=5, nRecons=5, nPagRecons=0, dim=10)

        self.widget = ScanSelectorWidget(parent=None)

    def tearDown(self):
        shutil.rmtree(self._folder1)
        shutil.rmtree(self._folder2)
        shutil.rmtree(self._folder3)

    def test(self):
        self.widget.add(self._folder1)
        self.widget.add(self._folder2)
        self.widget.add(self._folder3)
        self.assertEqual(self.widget.dataList.n_data(), 3)
        self.widget.remove(self._folder3)
        self.assertEqual(self.widget.dataList.n_data(), 2)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/08/2020"


import datetime
import logging
import os
import typing
from contextlib import AbstractContextManager

from nabu.pipeline.config import generate_nabu_configfile, parse_nabu_config_file
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from silx.utils.enum import Enum as _Enum

import tomwer.version
from tomwer.core.process.reconstruction.nabu.plane import NabuPlane
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.volume.edfvolume import EDFVolume
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.core.volume.jp2kvolume import JP2KVolume
from tomwer.core.volume.rawvolume import RawVolume
from tomwer.core.volume.tiffvolume import TIFFVolume

_logger = logging.getLogger(__name__)


class TomwerInfo(AbstractContextManager):
    """Simple context manager to add tomwer metadata to a dict before
    writing it"""

    def __init__(self, config_dict):
        self.config = config_dict

    def __enter__(self):
        self.config["other"] = {
            "tomwer_version": tomwer.version.version,
            "date": datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S"),
        }
        return self.config

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.config["other"]["tomwer_version"]
        del self.config["other"]["date"]


def retrieve_lst_of_value_from_str(my_string: str, type_) -> tuple:
    """
    Return a list of value from a string like '12,23' or '(12, 23)',
    '[12;23]', '12;23' or with the pattern from:to:step like '0:10:1'

    :param str mystring:
    :return: list of single value
    """
    if not isinstance(my_string, str):
        raise TypeError(
            f"my_string is expected to be a string. {type(my_string)} provided"
        )
    res = []
    my_string = my_string.replace("(", "")
    my_string = my_string.replace(")", "")
    my_string = my_string.replace("[", "")
    my_string = my_string.replace("]", "")
    if my_string.count(":") == 2:
        _from, _to, _step = my_string.split(":")
        _from, _to, _step = float(_from), float(_to), float(_step)
        if _from > _to:
            tmp = _to
            _to = _from
            _from = tmp
        while _from <= _to:
            res.append(_from)
            _from += _step
        return tuple(res)
    else:
        vals = my_string.replace(" ", "")
        vals = vals.replace("_", "")
        vals = vals.replace(";", ",").split(",")
        for val in vals:
            try:
                res.append(type_(val))
            except Exception:
                pass
        return tuple(res)


def get_nabu_resources_desc(scan: TomwerScanBase, method, workers=1) -> dict:
    """
    Create the descriptor of nabu's resources

    :param TomwerScanBase scan:
    :param str method:
    :return: nabu's description of resources to be used
    """
    assert isinstance(scan, TomwerScanBase)
    res = {
        "method": method,
        "cpu_workers": workers,
        "partition": "gpu",
        "memory_per_node": "90%",
        "threads_per_node": "100%",
    }
    return res


def get_nabu_about_desc(overwrite) -> dict:
    """
    Create the description for nabu's about

    :param self:
    :return:
    """
    return {"overwrite_results": str(bool(overwrite))}


def get_recons_volume_identifier(
    file_prefix: str,
    location: str,
    file_format: str,
    scan: TomwerScanBase,
    slice_index: typing.Union[int, None],
    axis: NabuPlane,
) -> tuple:
    """
    return tuple of DataUrl for existings slices
    """
    axis = NabuPlane.from_value(axis)
    file_format = file_format.lower()
    if file_format in ("hdf5", "h5", "hdf"):
        if slice_index is not None:
            # case of a single hdf5 file
            file_name = "_".join(
                (file_prefix, "plane", axis.value, str(slice_index).zfill(6))
            )
        else:
            file_name = file_prefix
        file_name = ".".join((file_name, file_format))
        file_path = os.path.join(location, file_name)

        if isinstance(scan, NXtomoScan):
            entry = scan.entry
        elif isinstance(scan, EDFTomoScan):
            entry = "entry"
        else:
            raise NotImplementedError(f"unrecognized scan type ({type(scan)})")

        volumes = (
            HDF5Volume(
                file_path=file_path,
                data_path="/".join([entry, "reconstruction"]),
            ),
        )
    elif file_format in ("vol", "raw"):
        if slice_index is not None:
            # case of a single hdf5 file
            file_name = "_".join(
                (file_prefix, "plane", axis.value, str(slice_index).zfill(6))
            )
        else:
            file_name = file_prefix
        file_name = ".".join((file_name, file_format))
        file_path = os.path.join(location, file_name)

        volumes = (RawVolume(file_path=file_path),)
    elif file_format in ("jp2", "jp2k", "edf", "tiff"):
        if file_format in ("jp2k", "jp2"):
            constructor = JP2KVolume
        elif file_format == "edf":
            constructor = EDFVolume
        elif file_format == "tiff":
            constructor = TIFFVolume
        else:
            raise NotImplementedError
        basename = file_prefix
        file_path = location
        volumes = (
            constructor(
                folder=location,
                volume_basename=basename,
            ),
        )

    # case of the multitiff. Not handled by tomwer
    # elif file_format == "tiff":
    #     # for single frame tiff nabu uses another convention by saving it 'directly'.
    #     volumes = (
    #         MultiTIFFVolume(
    #             file_path=os.path.join(
    #                 location,
    #                 file_prefix,
    #                 ".".join([file_prefix, file_format]),
    #             ),
    #         ),
    #     )

    else:
        raise ValueError(f"file format not managed: {file_format}")

    return tuple([volume.get_identifier() for volume in volumes])


def get_multi_cor_recons_volume_identifiers(
    scan: TomwerScanBase,
    slice_index: int,
    location: str,
    file_prefix: str,
    cors: tuple,
    file_format: str,
    axis=NabuPlane.XY,
) -> dict:
    """
    util to retrieve Volumes (identifier) reconstructed by nabu-multicor

    :param TomwerScanBase scan: scam processed by the nabu-multicor
    :param str location: location of the files
    :param tuple cors: cors for which we want the reconstructed slices. As this extension is created by nabu
                       the cor reference is in absolute.
    :param str file_format: file format of the reconstruction

    :return: a dict with absolute cor value as key and the Volume identifier as value
    """
    _logger.info("Deduce volume identifiers for nabu-multicor")
    if isinstance(slice_index, str):
        slice_index = slice_index_to_int(
            slice_index=slice_index,
            scan=scan,
            axis=axis,  # for now we always expect the multicor to be done along Z
        )
    assert isinstance(
        slice_index, int
    ), "slice_index is expected to be an int or to be converted to it"
    res = {}
    if isinstance(scan, EDFTomoScan):
        entry = "entry"
    else:
        entry = scan.entry

    for cor in cors:
        file_path = os.path.join(
            location,
            f"{file_prefix}_{cor:.3f}_{str(slice_index).zfill(5)}.{file_format}",
        )

        if file_format in ("hdf5", "h5", "hdf"):
            file_path = os.path.join(
                location,
                f"{file_prefix}_{cor:.3f}_{str(slice_index).zfill(5)}.{file_format}",
            )
            volume = HDF5Volume(
                file_path=file_path,
                data_path="/".join([entry, "reconstruction"]),
            )
        elif file_format in ("vol", "raw"):
            volume = (RawVolume(file_path=file_path),)
        elif file_format in ("jp2", "jp2k", "edf", "tiff"):
            if file_format in ("jp2k", "jp2"):
                constructor = JP2KVolume
            elif file_format == "edf":
                constructor = EDFVolume
            elif file_format == "tiff":
                constructor = TIFFVolume
            else:
                raise NotImplementedError
            file_path = location
            volume = constructor(
                folder=os.path.dirname(file_path),
                volume_basename=os.path.basename(file_path),
            )
        else:
            raise ValueError(f"file_format {file_format} is not handled for now")
        res[cor] = volume.get_identifier()
    return res


class _NabuMode(_Enum):
    FULL_FIELD = "standard acquisition"
    HALF_ACQ = "half acquisition"
    HELICAL = "helical acquisition"


class _NabuStages(_Enum):
    INI = "initialization"
    PRE = "pre-processing"
    PHASE = "phase"
    PROC = "processing"
    POST = "post-processing"
    VOLUME = "volume"

    @staticmethod
    def getStagesOrder():
        return (
            _NabuStages.INI,
            _NabuStages.PRE,
            _NabuStages.PHASE,
            _NabuStages.PROC,
            _NabuStages.POST,
        )

    @staticmethod
    def getProcessEnum(stage):
        """Return the process Enum associated to the stage"""
        stage = _NabuStages.from_value(stage)
        if stage is _NabuStages.INI:
            raise NotImplementedError()
        elif stage is _NabuStages.PRE:
            return _NabuPreprocessing
        elif stage is _NabuStages.PHASE:
            return _NabuPhase
        elif stage is _NabuStages.PROC:
            return _NabuProcessing
        elif stage is _NabuStages.POST:
            return _NabuPostProcessing
        raise NotImplementedError()


class _NabuPreprocessing(_Enum):
    """Define all the preprocessing action possible and the order they
    are applied on"""

    FLAT_FIELD_NORMALIZATION = "flat field normalization"
    CCD_FILTER = "hot spot correction"

    @staticmethod
    def getPreProcessOrder():
        return (
            _NabuPreprocessing.FLAT_FIELD_NORMALIZATION,
            _NabuPreprocessing.CCD_FILTER,
        )


class _NabuPhase(_Enum):
    """Define all the phase action possible and the order they
    are applied on"""

    PHASE = "phase retrieval"
    UNSHARP_MASK = "unsharp mask"
    LOGARITHM = "logarithm"

    @staticmethod
    def getPreProcessOrder():
        return (_NabuPhase.PHASE, _NabuPhase.UNSHARP_MASK, _NabuPhase.LOGARITHM)


class _NabuProcessing(_Enum):
    """Define all the processing action possible"""

    RECONSTRUCTION = "reconstruction"

    @staticmethod
    def getProcessOrder():
        return (_NabuProcessing.RECONSTRUCTION,)


class _NabuPostProcessing(_Enum):
    """Define all the post processing action available"""

    SAVE_DATA = "save"

    @staticmethod
    def getProcessOrder():
        return (_NabuPostProcessing.SAVE_DATA,)


class _NabuReconstructionMethods(_Enum):
    FBP = "FBP"


class _NabuPhaseMethod(_Enum):
    """
    Nabu phase method
    """

    PAGANIN = "Paganin"
    CTF = "CTF"
    NONE = "None"

    @classmethod
    def from_value(cls, value):
        if value in (None, ""):
            return _NabuPhaseMethod.NONE
        elif isinstance(value, str):
            if value.lower() == "paganin":
                return _NabuPhaseMethod.PAGANIN
            elif value.lower() == "none":
                return _NabuPhaseMethod.NONE
            elif value.lower() == "ctf":
                return _NabuPhaseMethod.CTF
        else:
            return super().from_value(value=value)


class _NabuFBPFilterType(_Enum):
    RAMLAK = "ramlak"
    SHEPP_LOGAN = "shepp-logan"
    COSINE = "cosine"
    HAMMING = "hamming"
    HANN = "hann"
    TUKEY = "tukey"
    LANCZOS = "lanczos"
    HILBERT = "hilbert"


class _NabuPaddingType(_Enum):
    ZEROS = "zeros"
    EDGES = "edges"


class RingCorrectionMethod(_Enum):
    NONE = "None"
    MUNCH = "munch"
    VO = "vo"
    MEAN_SUBTRACTION = "mean-subtraction"
    MEAN_DIVISION = "mean-division"


def nabu_std_err_has_error(errs: typing.Optional[bytes]):
    """
    small util to parse stderr where some warning can exists.
    But I don't think we want to catch all warnings from nabu so this is a (bad) concession
    This will disapear when execution will be done directly from a tomwer thread instead of a subprocess
    """

    def ignore(line) -> bool:
        return (
            "warning" in line
            or "Warning" in line
            or "cupyx.jit.rawkernel" in line
            or "deprecated" in line
            or line.replace(" ", "") == ""
            or "unable to load" in line
            or "deprecated" in line
            or "self.module = SourceModule(self.src, **self.sourcemodule_kwargs)"
            in line
            or "return SourceModule(" in line
            or "CUBLAS" in line
            or "Not supported for EDF"
            in line  # debatable but very disturbing from the gui side... anyway EDF days are coming to an end
            or "PerformanceWarning" in line
            or "jitify._init_module()" in line
            or " unable to import 'siphash24.siphash13" in line
            or "_create_built_program_from_source_cached" in line
            or "prg.build(options_bytes," in line
        )

    if errs is None:
        return False
    else:
        for line in errs.decode("UTF-8").split("\n"):
            if not ignore(line):
                return True
    return False


def update_cfg_file_after_transfer(config_file_path, old_path, new_path):
    """
    update nabu configuration file path from /lbsram/data to /data
    """
    if old_path is None or new_path is None:
        return

    # load configucation file
    config_as_dict = parse_nabu_config_file(config_file_path)
    assert isinstance(config_as_dict, dict)

    # update paths
    paths_to_update = (
        ("dataset", "location"),
        ("output", "location"),
        ("pipeline", "steps_file"),
    )
    for section, field in paths_to_update:
        # update dataset location and output location
        if section in config_as_dict:
            location = config_as_dict[section].get(field, None)
            if location is not None:
                config_as_dict[section][field] = location.replace(old_path, new_path, 1)
    # overwrite file
    generate_nabu_configfile(
        fname=config_file_path,
        default_config=nabu_fullfield_default_config,
        config=config_as_dict,
        options_level="advanced",
    )


def slice_index_to_int(
    slice_index: typing.Union[int, str],
    scan: TomwerScanBase,
    axis=NabuPlane.XY,
) -> int:
    """
    cast a slice to an index. The slice can be a string in ['first', 'last', 'middle']
    """
    axis = NabuPlane.from_value(axis)
    if slice_index == "fisrt":
        return 0
    elif slice_index == "last":
        if scan is None:
            # backward compatibility in the case the scan is not provided. Should not happen anymore
            _logger.warning("Scan not provided. Consider the 2048 width detector")
            return 2047
        elif scan.dim_2 is not None:
            return scan.dim_2 - 1
        else:
            # this could happen on some EDF scans. Not expected to happen for HDF5
            _logger.warning("unable to get dim size, guess this is 2048 width")
            # in this
            return 2047
    elif slice_index == "middle":
        if scan is None:
            # backward compatibility in the case the scan is not provided. Should not happen anymore
            _logger.warning("Scan not provided. Consider the 1024 width detector")
            # default middle.
            return 1024
        elif axis is NabuPlane.XY:
            if scan.dim_2 is None:
                _logger.warning("unable to get dim size, guess this is 2048 height")
                return 1024
            else:
                return scan.dim_2 // 2
        elif axis in (NabuPlane.YZ, NabuPlane.XZ):
            if scan.dim_1 is None:
                _logger.warning("unable to get dim size, guess this is 2048 width")
                return 1024
            else:
                return scan.dim_1 // 2
        else:
            raise ValueError(f"axis {axis} is not handled")
    else:
        return int(slice_index)


def get_nabu_multicor_file_prefix(scan):
    if isinstance(scan, EDFTomoScan):
        dataset_path = scan.path
    elif isinstance(scan, NXtomoScan):
        dataset_path = scan.master_file
    else:
        raise TypeError(f"{type(scan)} is not handled")

    if os.path.isfile(dataset_path):  # hdf5
        file_prefix = os.path.basename(dataset_path).split(".")[0]
    elif os.path.isdir(dataset_path):
        file_prefix = os.path.basename(dataset_path)
    else:
        raise ValueError(f"dataset location {scan.path} is neither a file or directory")
    file_prefix += "_rec"  # avoid overwriting dataset
    return file_prefix



import argparse
import logging
import os
import shutil
import signal
import sys
import silx
from typing import List
from contextlib import closing
from logging.handlers import RotatingFileHandler
from urllib.request import urlopen
from xml.sax.saxutils import escape

import pyqtgraph
from orangecanvas.application.canvasmain import DockWidget
from orangecanvas.application.outputview import (
    ExceptHook,
    TerminalTextDocument,
    TextStream,
)
from orangecanvas.document.usagestatistics import UsageStatistics
from orangecanvas.main import Main as ocMain
from orangewidget.workflow import config
from orangewidget.workflow.config import data_dir_base
from orangewidget.workflow.errorreporting import handle_exception
from orangewidget.workflow.mainwindow import OWCanvasMainWindow as _MainWindow
from processview.gui.processmanager import ProcessManagerWindow
from silx.gui import qt

import tomwer.version
from tomwer.core.log.logger import TomwerLogger
from orangecanvas.preview import previewdialog, previewmodel
from orangecanvas.application import examples as orange_examples
import ewoks
import tomoscan.version

from .config import TomwerConfig, TomwerSplashScreen

try:
    import nabu
except ImportError:
    has_nabu = False
else:
    has_nabu = True
try:
    import nxtomomill.version
except ImportError:
    has_nxtomomill = False
else:
    has_nxtomomill = True
try:
    import nxtomo
except ImportError:
    has_nxotmo = False
else:
    has_nxotmo = True
try:
    import sluurp
except ImportError:
    has_sluurp = False
else:
    has_sluurp = True

_logger = logging.getLogger(__file__)

MAX_LOG_FILE = 10
"""Maximal log file kepts for orange"""

LOG_FILE_NAME = "tomwer.log"

LOG_FOLDER = "/var/log/tomwer"


class MainWindow(_MainWindow):
    HELPDESK_URL = "https://requests.esrf.fr/plugins/servlet/desk/portal/41"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.process_supervisor_dock = DockWidget(
            self.tr("object supervisor"),
            self,
            objectName="processes-dock",
            allowedAreas=qt.Qt.BottomDockWidgetArea,
            visible=self.show_processes_manager_action.isChecked(),
        )

        self.process_supervisor_dock.setWidget(ProcessManagerWindow(parent=None))
        self.process_supervisor_dock.visibilityChanged[bool].connect(
            self.show_processes_manager_action.setChecked
        )
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self.process_supervisor_dock)

    def setup_actions(self):
        super().setup_actions()
        # create the action to connect with it
        self.show_processes_manager_action = qt.QAction(
            self.tr("&object supervisor"),
            self,
            toolTip=self.tr("Show object states relative to processes."),
            checkable=True,
            triggered=lambda checked: self.process_supervisor_dock.setVisible(checked),
        )

    def setup_menu(self):
        super().setup_menu()
        self.view_menu.addAction(self.show_processes_manager_action)

        self.simple_example_action = qt.QAction(
            self.tr("simple use cases"),
            self,
            objectName="simple-use-cases",
            toolTip=self.tr("Some simple examples."),
            triggered=self.open_simple_examples,
            menuRole=qt.QAction.AboutRole,
        )

        self.dark_flat_example_action = qt.QAction(
            self.tr("darks and flats manipulations"),
            self,
            objectName="dark-flat-manipulation",
            toolTip=self.tr("Examples around dark and flat."),
            triggered=self.open_dark_flat_examples,
            menuRole=qt.QAction.AboutRole,
        )
        self.slurm_execution_example_action = qt.QAction(
            self.tr("remote processing with slurm"),
            self,
            objectName="execution-on-slurm",
            toolTip=self.tr("Examples of some execution over slurm."),
            triggered=self.open_slurm_execution_example,
            menuRole=qt.QAction.AboutRole,
        )
        self.cor_search_example_action = qt.QAction(
            self.tr("cor search advance example"),
            self,
            objectName="cor-search-example",
            toolTip=self.tr("Examples of ways to search for the Center of rotation."),
            triggered=self.open_cor_search_example,
            menuRole=qt.QAction.AboutRole,
        )
        self.python_script_example_action = qt.QAction(
            self.tr("python script examples"),
            self,
            objectName="python-script-example",
            toolTip=self.tr("Examples of some python script usage."),
            triggered=self.open_python_script_example,
            menuRole=qt.QAction.AboutRole,
        )
        # Predefined workflows menu.
        self.predefined_workflows_menu = qt.QMenu(
            self.tr("&Examples"),
            self.menuBar(),
            objectName="examples-menu",
        )
        self.predefined_workflows_menu.addActions(
            [
                self.simple_example_action,
                self.cor_search_example_action,
                self.dark_flat_example_action,
                self.slurm_execution_example_action,
                self.python_script_example_action,
            ]
        )

        self.menuBar().addMenu(self.predefined_workflows_menu)

        self._helpdeskButton = qt.QAction(
            self.tr("&Helpdesk"),
            self,
            objectName="helpdeskButton",
            triggered=self._contactUs,
            toolTip="contact esrf data processing helpdesk if you have some troubles with the application",
        )
        self.menuBar().addAction(self._helpdeskButton)

    def _contactUs(self, *args, **kwargs):
        try:
            qt.QDesktopServices.openUrl(qt.QUrl(self.HELPDESK_URL))
        except Exception as e:
            _logger.error(
                f"Failed to launch helpdesk web page ({self.HELPDESK_URL}). Error is {e}"
            )

    def __open_examples(self, examples, filter_names=None):
        """
        filter the existing tutorials by names contained in the file name
        """
        items = [previewmodel.PreviewItem(path=t.abspath()) for t in examples]

        def my_filter(value):
            for filter_name in filter_names:
                if filter_name in value:
                    return True
            return False

        if filter_names is not None:
            items = filter(
                lambda item: my_filter(os.path.basename(item.path())),
                items,
            )
        dialog = previewdialog.PreviewDialog(self)
        model = previewmodel.PreviewModel(dialog, items=items)
        title = self.tr("Example Workflows")
        dialog.setWindowTitle(title)
        template = '<h3 style="font-size: 26px">\n' "{0}\n" "</h3>"

        dialog.setHeading(template.format(title))
        dialog.setModel(model)

        model.delayedScanUpdate()
        status = dialog.exec()
        index = dialog.currentIndex()

        dialog.deleteLater()

        if status == qt.QDialog.Accepted:
            selected = model.item(index)
            self.open_example_scheme(selected.path())
        return status

    def open_about(self):
        dlg = AboutDialog(self)
        dlg.setAttribute(qt.Qt.WA_DeleteOnClose)
        dlg.exec()

    def open_simple_examples(self):
        return self.__open_examples(
            orange_examples.workflows(), filter_names=("simple", "EBS_tomo_listener")
        )

    def open_slurm_execution_example(self):
        return self.__open_examples(
            orange_examples.workflows(), filter_names=("slurm",)
        )

    def open_dark_flat_examples(self):
        return self.__open_examples(
            orange_examples.workflows(), filter_names=("darks",)
        )

    def open_python_script_example(self):
        return self.__open_examples(
            orange_examples.workflows(), filter_names=("_script",)
        )

    def open_cor_search_example(self):
        return self.__open_examples(
            orange_examples.workflows(), filter_names=("find_cor", "cor_search")
        )


log = logging.getLogger(__name__)


def check_for_updates() -> bool:
    return False


def send_usage_statistics() -> bool:
    return False


def pull_notifications() -> bool:
    return False


def make_sql_logger(level=logging.INFO):
    sql_log = logging.getLogger("sql_log")
    sql_log.setLevel(level)
    handler = RotatingFileHandler(
        os.path.join(config.log_dir(), "sql.log"), maxBytes=1e7, backupCount=2
    )
    sql_log.addHandler(handler)


class _OMain(ocMain):
    DefaultConfig = "tomwer.app.canvas_launcher.launcher.TomwerConfig"

    def run(self, argv):
        # Allow termination with CTRL + C
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        # Disable pyqtgraph's atexit and QApplication.aboutToQuit cleanup handlers.
        pyqtgraph.setConfigOption("exitCleanup", False)
        super().run(argv)

    def argument_parser(self) -> argparse.ArgumentParser:
        parser = super().argument_parser()
        parser.add_argument(
            "--use-opengl-plot",
            help="Use OpenGL for plots (instead of matplotlib)",
            action="store_true",
            default=False,
        )
        return parser

    def parse_arguments(self, argv: List[str]):
        super().parse_arguments(argv)
        # define silx plot backend. For now the si;pler is to
        # define an environment variable.
        # this way we should also be able to let the user define it
        # on it's own later...

        if self.options.use_opengl_plot:

            # other standalones are not passing by the silx.config
            # but for the canvas this is way simpler to use it.
            # then all plot which have backend==None will pick the default plot backend
            silx.config.DEFAULT_PLOT_BACKEND = "gl"
        else:
            silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"

    def setup_logging(self):
        super().setup_logging()
        make_sql_logger(self.options.log_level)

    def setup_application(self):
        super().setup_application()
        # NOTE: No OWWidgetBase subclass should be imported before this

        self._update_check = check_for_updates()
        self._send_stat = send_usage_statistics()
        self._pull_notifs = pull_notifications()

        settings = qt.QSettings()
        settings.setValue(
            "startup/launch-count",
            settings.value("startup/launch-count", 0, int) + 1,
        )

        UsageStatistics.set_enabled(False)

    def show_splash_message(self, message: str, color=qt.QColor("#FFD39F")):
        super().show_splash_message(message, color)

    def create_main_window(self):
        window = MainWindow()
        return window

    def setup_sys_redirections(self):
        super().setup_sys_redirections()
        if isinstance(sys.excepthook, ExceptHook):
            sys.excepthook.handledException.connect(handle_exception)

    def tear_down_sys_redirections(self):
        if isinstance(sys.excepthook, ExceptHook):
            sys.excepthook.handledException.disconnect(handle_exception)
        super().tear_down_sys_redirections()

    def splash_screen(self):
        """Return the application splash screen"""
        settings = qt.QSettings()
        options = self.options
        want_splash = (
            settings.value("startup/show-splash-screen", True, type=bool)
            and not options.no_splash
        )

        if want_splash:
            pm, rect = self.config.splash_screen()
            splash_screen = TomwerSplashScreen(pixmap=pm, textRect=rect)
            splash_screen.setAttribute(qt.Qt.WA_DeleteOnClose)
            splash_screen.setFont(qt.QFont("Helvetica", 12))
            palette = splash_screen.palette()
            color = qt.QColor("#b3baba")
            palette.setColor(qt.QPalette.Text, color)
            splash_screen.setPalette(palette)
        else:
            splash_screen = None
        return splash_screen


def data_dir():
    return os.path.join(data_dir_base(), "tomwer", tomwer.version.version)


def widget_settings_dir():
    return os.path.join(data_dir(), "widgets")


ABOUT_TEMPLATE = """\
<h4>{name}</h4>
<p>tomwer version: {tomwer_version}</p>
<p>nabu version: {nabu_version}</p>
<p>nxtomomill version: {nxtomomill_version}</p>
<p>nxtomo version: {nxtomo_version}</p>
<p>tomoscan version: {tomoscan_version}</p>
<p>sluurp version: {sluurp_version}</p>
<p>ewoks version: {ewoks_version}</p>
"""


class AboutDialog(qt.QDialog):
    def __init__(self, parent=None, **kwargs) -> None:
        super().__init__(parent, **kwargs)
        layout = qt.QVBoxLayout()
        label = qt.QLabel(self)

        pixmap, _ = TomwerConfig.splash_screen()
        pixmap = pixmap.scaled(150, 150)

        label.setPixmap(pixmap)

        layout.addWidget(label, qt.Qt.AlignCenter)

        text = ABOUT_TEMPLATE.format(
            name=escape("tomwer"),
            tomwer_version=escape(tomwer.version.version),
            nabu_version=escape(nabu.version if has_nabu else "not installed"),
            nxtomomill_version=escape(
                nxtomomill.version.version if has_nxtomomill else "not installed"
            ),
            nxtomo_version=escape(
                nxtomo.__version__ if has_nxotmo else "not installed"
            ),
            tomoscan_version=escape(tomoscan.version.version),
            sluurp_version=escape(
                sluurp.__version__ if has_sluurp else "not installed"
            ),
            ewoks_version=escape(
                ewoks.__version__ if hasattr(ewoks, "__version__") else "???"
            ),
        )
        text_label = qt.QLabel(text)
        layout.addWidget(text_label, qt.Qt.AlignCenter)

        buttons = qt.QDialogButtonBox(qt.QDialogButtonBox.Close, qt.Qt.Horizontal, self)
        layout.addWidget(buttons)
        buttons.rejected.connect(self.accept)
        layout.setSizeConstraint(qt.QVBoxLayout.SetFixedSize)
        self.setLayout(layout)


class OMain(_OMain):
    config: TomwerConfig
    DefaultConfig = "tomwer.app.canvas_launcher.config.TomwerConfig"

    def run(self, argv):
        log.info("Clearing widget settings")
        shutil.rmtree(widget_settings_dir(), ignore_errors=True)
        dealWithLogFile()
        super().run(argv)

    def setup_application(self):
        qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
        return super().setup_application()

    def setup_logging(self):
        super().setup_logging()
        rootlogger = logging.getLogger()
        rootlogger = TomwerLogger(rootlogger)
        logging.setLoggerClass(TomwerLogger)

    def setup_sys_redirections(self):
        self._tomwLogger = TomwerLogger("tomwer")
        try:
            self.output = doc = TerminalTextDocument()

            stdout = TextStream(objectName="-stdout")
            stderr = TextStream(objectName="-stderr")
            doc.connectStream(stdout)
            doc.connectStream(stderr, color=qt.Qt.red)

            if sys.stdout is not None:
                stdout.stream.connect(sys.stdout.write, qt.Qt.DirectConnection)

            self.__stdout__ = sys.stdout
            sys.stdout = stdout

            if sys.stderr is not None:
                stderr.stream.connect(sys.stderr.write, qt.Qt.DirectConnection)

            self.__stderr__ = sys.stderr
            sys.stderr = stderr
            self.__excepthook__ = sys.excepthook
            sys.excepthook = ExceptHook(stream=stderr)

            self.stack.push(closing(stdout))
            self.stack.push(closing(stderr))
        except Exception:
            super().setup_sys_redirections()

    def argument_parser(self) -> argparse.ArgumentParser:
        parser = super().argument_parser()
        for action in parser._actions:
            if action.dest == "clear_widget_settings":
                parser._remove_action(action)
                break

        parser.add_argument(
            "--no-color-stdout-logs",
            "--no-colored-logs",
            action="store_true",
            help="instead of having logs in the log view, color logs of the stdout",
            default=False,
        )
        return parser

    def create_main_window(self):
        window = MainWindow()
        return window


def dealWithLogFile():
    """Move log file history across log file hierarchy and create the new log file"""

    # move log file if exists
    for i in range(MAX_LOG_FILE):
        logFile = LOG_FILE_NAME
        if os.path.exists(LOG_FOLDER) and os.access(LOG_FOLDER, os.W_OK):
            logFile = os.path.join(LOG_FOLDER, logFile)
        defLogName = logFile

        iLog = MAX_LOG_FILE - i
        maxLogNameN1 = logFile + "." + str(iLog)
        if iLog - 1 == 0:
            maxLogNameN2 = defLogName
        else:
            maxLogNameN2 = logFile + "." + str(iLog - 1)
        if os.path.exists(maxLogNameN2):
            try:
                stat = os.stat(maxLogNameN2)
                shutil.copy(maxLogNameN2, maxLogNameN1)
                os.utime(maxLogNameN1, (stat.st_atime, stat.st_mtime))
            except Exception:
                pass
    # create a new log file
    if os.path.exists(LOG_FOLDER) and os.access(LOG_FOLDER, os.W_OK):
        logFile = os.path.join(LOG_FOLDER, logFile)
        logging.basicConfig(
            filename=logFile,
            filemode="w",
            level=logging.WARNING,
            format="%(asctime)s %(message)s",
        )


def check_is_latest_release() -> bool:
    """Check if the current version is the latest release."""
    url = "https://gitlab.esrf.fr/tomotools/tomwer/-/raw/master/tomwer/version.py"
    current_version = tomwer.version.version
    try:
        version_file_html = urlopen(url, data=None, timeout=10)
    except Exception as e:
        _logger.warning(f"Fail to load version of the latest release. Reason is {e}")
        return True
    else:
        latest_release_version = None
        for line in version_file_html.readlines():
            t_line = line.decode("utf-8")
            t_line = t_line.replace(" ", "")
            if t_line.startswith("latest_release_version_info="):
                latest_release_version = t_line.replace(
                    "latest_release_version_info=", ""
                )
                break
        if latest_release_version is None:
            _logger.warning("Unable to find the version of the latest " "release.")

        elif current_version < latest_release_version:
            msg = qt.QMessageBox()
            msg.setIcon(qt.QMessageBox.Question)
            types = qt.QMessageBox.Ok | qt.QMessageBox.Cancel
            message = (
                "The version you want to use ({}) is not the latest "
                "version ({}). Do you want to continue ?"
            )
            msg.setStandardButtons(types)
            msg.setWindowTitle("No the latest version")
            msg.setText(message)
            return msg.exec_() == qt.QMessageBox.Ok
        return True



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""some utils relative to PyHST
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "17/02/2021"


import numpy
from silx.gui import qt


class LogSlider(qt.QWidget):
    """Slider to select a value with a QSlider displayed with log scale"""

    valueChanged = qt.Signal(float)
    """signal emitted when the value changed"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        # QSlider
        self._slider = qt.QSlider(self)
        self._slider.setOrientation(qt.Qt.Horizontal)
        self.layout().addWidget(self._slider, 0, 0, 1, 1)
        # Double spin box
        self._valueQBSB = qt.QDoubleSpinBox(self)
        self.layout().addWidget(self._valueQBSB, 0, 1, 1, 1)

        # connect signal / slot
        self._slider.valueChanged.connect(self._sliderValueChanged)
        self._valueQBSB.valueChanged.connect(self._qdsbValueChanged)
        # set up
        self.setRange(1, 100)
        self.setValue(5)

    def setSuffix(self, txt):
        self._valueQBSB.setSuffix(txt)

    def setPrefix(self, txt):
        self._valueQBSB.setPrefix(txt)

    def setRange(self, min_: float, max_: float) -> None:
        """
        Define slider range

        :param float min_:
        :param float max_:
        """
        if min_ <= 0.0 or max_ <= 0.0:
            raise ValueError("LogSlider can only handled positive values")
        self._valueQBSB.setRange(min_, max_)
        self._slider.setRange(int(numpy.log(min_)), int(numpy.log(max_)))

    def _sliderValueChanged(self, *args, **kwargs):
        old = self._valueQBSB.blockSignals(True)
        self._valueQBSB.setValue(numpy.exp(self._slider.value()))
        self._valueQBSB.blockSignals(old)
        self.valueChanged.emit(self.value())

    def _qdsbValueChanged(self, *args, **kwargs):
        old = self._slider.blockSignals(True)
        self._slider.setValue(int(numpy.log(self._valueQBSB.value())))
        self._slider.blockSignals(old)
        self.valueChanged.emit(self.value())

    def value(self):
        return self._valueQBSB.value()

    def setValue(self, value):
        self._valueQBSB.setValue(value)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/08/2020"


import typing
from asyncio.log import logger

from silx.gui import qt

from tomwer.core.process.reconstruction.nabu.utils import (
    _NabuStages,
)
from tomwer.gui import icons
from tomwer.gui.configuration.action import (
    BasicConfigurationAction,
    ExpertConfigurationAction,
    MinimalisticConfigurationAction,
)
from tomwer.gui.configuration.level import ConfigurationLevel
from tomwer.gui.reconstruction.nabu.nabuconfig.base import _NabuStageConfigBase
from tomwer.gui.reconstruction.nabu.nabuconfig.output import (
    NabuOutputLocationWidget,
    QNabuFileFormatComboBox,
)


class SliceSelector(qt.QWidget):
    """
    Widget to select a value for defining a ROI to reconstruction (start or
    end)
    """

    valueChanged = qt.Signal()

    def __init__(
        self, label, auto_alias, is_end, grid_layout, index_layout, parent=None
    ):
        qt.QWidget.__init__(self, parent=parent)
        self._label = qt.QLabel(label, parent=self)
        grid_layout.addWidget(self._label, index_layout, 0, 1, 1)
        self._qcb = qt.QCheckBox(auto_alias)
        grid_layout.addWidget(self._qcb, index_layout, 1, 1, 1)
        grid_layout.addWidget(qt.QLabel("or", self), index_layout, 2, 1, 1)
        self._qle = qt.QLineEdit("0", self)
        self._validator = qt.QIntValidator()
        self._qle.setValidator(self._validator)
        grid_layout.addWidget(self._qle, index_layout, 3, 1, 1)
        self._is_end = is_end
        if is_end is True:
            self._qle.setToolTip("-1 will be the last slice")
            self._validator.setBottom(-1)
        else:
            self._validator.setBottom(0)

        # set up
        self._qcb.setChecked(True)
        self._qle.setEnabled(False)
        if is_end:
            self._qle.setText("-1")

        # connect signal / slot
        self._qcb.toggled.connect(self._qle.setDisabled)
        self._qcb.toggled.connect(self._triggerValueChanged)
        self._qle.editingFinished.connect(self._triggerValueChanged)

    def _triggerValueChanged(self, *args, **kwargs):
        self.valueChanged.emit()

    def setMaximum(self, maximum):
        self._validator.setTop(maximum)

    def value(self):
        if self._qcb.isChecked():
            return -1 if self._is_end else 0
        else:
            return int(self._qle.text())

    def setValue(self, value):
        self._qle.setText(str(value))
        if self._is_end:
            self._qcb.setChecked(value == -1)
        else:
            self._qcb.setChecked(value == 0)


class NabuVolumeWidget(_NabuStageConfigBase, qt.QWidget):
    """
    Widget dedicated to manage the volume reconstruction from nabu
    """

    sigConfigChanged = qt.Signal()
    """Signal emitted when the configuration change"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent=parent)
        _NabuStageConfigBase.__init__(self, stage=_NabuStages.VOLUME)
        self.setLayout(qt.QGridLayout())

        # warning about requires parameters from a nabu slice reconstruction
        self._warning_widget = qt.QWidget(parent=self)
        self._warning_widget.setLayout(qt.QHBoxLayout())
        self._warningIconL = qt.QLabel("", parent=self)
        self._warningIconL.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self._warningIconL.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self._warningIconR = qt.QLabel("", parent=self)
        self._warningIconR.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self._warningIconR.setAlignment(qt.Qt.AlignLeft | qt.Qt.AlignVCenter)
        warning_icon = icons.getQIcon("warning")
        self._warningIconL.setPixmap(warning_icon.pixmap(20, 20))
        self._warningIconR.setPixmap(warning_icon.pixmap(20, 20))
        self._warning_widget.layout().addWidget(self._warningIconL)

        self._labelWarning = qt.QLabel(
            "Requires parameters from nabu slice reconstruction.", parent=self
        )
        self._labelWarning.setToolTip(
            'This mean that a "nabu slice '
            'reconstruction" widget should exist upstream'
        )
        self._warning_widget.layout().addWidget(self._labelWarning)
        self._warning_widget.layout().addWidget(self._warningIconR)
        self.layout().addWidget(self._warning_widget, 0, 0, 1, 2)

        # volume from and to option
        self._groupBox = qt.QGroupBox("volume to reconstruct", parent=self)
        self._groupBox.setLayout(qt.QGridLayout())
        self.layout().addWidget(self._groupBox, 2, 0, 1, 2)

        self._fromSlice = SliceSelector(
            label="from slice",
            auto_alias="start",
            is_end=False,
            parent=self,
            grid_layout=self._groupBox.layout(),
            index_layout=0,
        )
        self._groupBox.layout().addWidget(self._fromSlice)
        self.registerWidget(self._groupBox, "required")

        self._toSlice = SliceSelector(
            label="to slice",
            auto_alias="end",
            is_end=True,
            parent=self,
            grid_layout=self._groupBox.layout(),
            index_layout=1,
        )
        self._groupBox.layout().addWidget(self._toSlice)

        # option to compute histogram
        self._histogramCB = qt.QCheckBox("compute volume histogram of values", self)
        self.layout().addWidget(self._histogramCB, 3, 0, 1, 2)
        self.registerWidget(self._histogramCB, "required")

        # output format can be redefined
        self._redefineNabuFileFormat = qt.QGroupBox(
            "change received file format from 'nabu slice'", self
        )
        self._redefineNabuFileFormat.setCheckable(True)
        self._redefineNabuFileFormat.setLayout(qt.QVBoxLayout())
        self._fileformatCB = QNabuFileFormatComboBox(self._redefineNabuFileFormat)
        self._redefineNabuFileFormat.layout().addWidget(self._fileformatCB)
        self.layout().addWidget(self._redefineNabuFileFormat, 4, 0, 1, 2)
        self.registerWidget(self._fileformatCB, "advanced")
        self.registerWidget(self._redefineNabuFileFormat, "advanced")

        # redefine output location
        self._redefineNabuOutputLocation = qt.QGroupBox(
            "change received reconstruction output location from 'nabu slice'", self
        )
        self._redefineNabuOutputLocation.setCheckable(True)
        self._redefineNabuOutputLocation.setLayout(qt.QVBoxLayout())
        self._outputLocationWidget = NabuOutputLocationWidget(parent=self)
        self._redefineNabuOutputLocation.layout().addWidget(self._outputLocationWidget)
        self.layout().addWidget(self._redefineNabuOutputLocation, 5, 0, 1, 2)
        self.registerWidget(self._redefineNabuOutputLocation, "advanced")
        self.registerWidget(self._outputLocationWidget, "advanced")

        # gpu percentage
        self._labelGPUFract = qt.QLabel("GPU percentage usage", self)
        self.layout().addWidget(self._labelGPUFract, 50, 0, 1, 1)
        self.registerWidget(self._labelGPUFract, "advanced")
        self._spbGPUFract = qt.QSpinBox(self)
        self._spbGPUFract.setSingleStep(10)
        self._spbGPUFract.setSuffix("%")
        self._spbGPUFract.setMinimum(1)
        self._spbGPUFract.setMaximum(100)
        self.layout().addWidget(self._spbGPUFract, 50, 1, 1, 1)
        self._spbGPUFract.setToolTip("Which fraction of GPU memory to use.")
        self.registerWidget(self._spbGPUFract, "advanced")

        # cpu percentage
        self._labelCPUFract = qt.QLabel("CPU percentage usage", self)
        self.layout().addWidget(self._labelCPUFract, 51, 0, 1, 1)
        self.registerWidget(self._labelCPUFract, "advanced")
        self._spbCPUFract = qt.QSpinBox(self)
        self._spbCPUFract.setSingleStep(10)
        self._spbCPUFract.setSuffix("%")
        self._spbCPUFract.setMinimum(1)
        self._spbCPUFract.setMaximum(100)
        self.layout().addWidget(self._spbCPUFract, 52, 1, 1, 1)
        self._spbCPUFract.setToolTip("Which fraction of CPU memory to use.")
        self.registerWidget(self._spbCPUFract, "advanced")

        # opt use phase paganin
        self._usePhaseQCB = qt.QCheckBox("Use phase paganin")
        self.layout().addWidget(self._usePhaseQCB, 53, 0, 1, 2)
        self.registerWidget(self._usePhaseQCB, "advanced")
        self._usePhaseQCB.setToolTip(
            "Whether to use a margin when " "performing phase retrieval."
        )

        # spacer for style
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 99, 0, 1, 1)

        # set up
        self._histogramCB.setChecked(True)
        self._spbGPUFract.setValue(90)
        self._spbCPUFract.setValue(90)
        self._usePhaseQCB.setChecked(True)
        self._redefineNabuFileFormat.setChecked(False)
        self._redefineNabuOutputLocation.setChecked(False)
        self.setConfigurationLevel("required")

        # connect signal / slot
        self._fromSlice.valueChanged.connect(self._triggerSigConfChanged)
        self._toSlice.valueChanged.connect(self._triggerSigConfChanged)
        self._spbGPUFract.valueChanged.connect(self._triggerSigConfChanged)
        self._spbCPUFract.valueChanged.connect(self._triggerSigConfChanged)
        self._usePhaseQCB.toggled.connect(self._triggerSigConfChanged)
        self._outputLocationWidget._outputDirQLE.editingFinished.connect(
            self._triggerSigConfChanged
        )
        self._outputLocationWidget.sigOutputChanged.connect(self._triggerSigConfChanged)
        self._redefineNabuOutputLocation.toggled.connect(self._triggerSigConfChanged)

    def redefineOutputLocation(self):
        return self._redefineNabuOutputLocation.isChecked()

    def setRedefineOutputLocation(self, redefine: bool):
        self._redefineNabuOutputLocation.setChecked(redefine)

    def getNabuOutputLocation(self):
        if self.redefineOutputLocation():
            return self._outputLocationWidget.getOutputDir()
        else:
            return None

    def setNabuOutputLocation(self, location: typing.Optional[str]):
        if location is None:
            self._redefineNabuOutputLocation.setChecked(False)
        else:
            self._redefineNabuOutputLocation.setChecked(True)
            self._outputLocationWidget.setOutputDirMode("other")
            self._outputLocationWidget.setOutputDir(location)

    def redefineNabuFileFormat(self):
        return self._redefineNabuFileFormat.isChecked()

    def getNabuFileFormat(self):
        if self.redefineNabuFileFormat():
            return self._fileformatCB.currentText()
        else:
            return None

    def setNabuFileFormat(self, file_format: typing.Optional[str]):
        if file_format is None:
            self._redefineNabuFileFormat.setChecked(False)
        else:
            assert isinstance(file_format, str)
            self._redefineNabuFileFormat.setChecked(True)
            idx = self._fileformatCB.findText(file_format)
            if idx < 0:
                logger.error(f"unable to find file format {file_format}")
            else:
                self._fileformatCB.setCurrentIndex(idx)

    def _triggerSigConfChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def getCPUFract(self):
        return self._spbCPUFract.value() / 100.0

    def setCPUFract(self, value):
        return self._spbCPUFract.setValue(int(value * 100.0))

    def getGPUFract(self):
        return self._spbGPUFract.value() / 100.0

    def setGPUFract(self, value):
        return self._spbGPUFract.setValue(int(value * 100.0))

    def getStartZ(self):
        return self._fromSlice.value()

    def setStartZ(self, value):
        self._fromSlice.setValue(value)

    def getEndZ(self):
        return self._toSlice.value()

    def setEndZ(self, value):
        self._toSlice.setValue(value)

    def phasePaganinRequested(self):
        return self._usePhaseQCB.isChecked()

    def setPhasePaganinRequested(self, checked):
        self._usePhaseQCB.setChecked(checked)

    def setScan(self, scan):
        if len(scan.projections) > 0:
            self._fromSlice.setMaximum(len(scan.projections))
        else:
            self._fromSlice.setMaximum(999999999)

    def isHistogramRequested(self):
        return self._histogramCB.isChecked()

    def setHistogramRequested(self, requested):
        self._histogramCB.setChecked(requested)

    def getConfiguration(self):
        return {
            "start_z": self.getStartZ(),
            "end_z": self.getEndZ(),
            "gpu_mem_fraction": self.getGPUFract(),
            "cpu_mem_fraction": self.getCPUFract(),
            "use_phase_margin": self.phasePaganinRequested(),
            "postproc": {
                "output_histogram": int(self.isHistogramRequested()),
            },
            "new_output_file_format": self.getNabuFileFormat() or "",
            "new_output_location": self.getNabuOutputLocation() or "",
            "output_dir_mode": self._outputLocationWidget.getOutputDirMode().value,
            "overwrite_output_location": self.redefineOutputLocation(),
        }

    def setConfiguration(self, config):
        if "start_z" in config:
            self.setStartZ(config["start_z"])
        if "end_z" in config:
            self.setEndZ(config["end_z"])
        if "gpu_mem_fraction" in config:
            self.setGPUFract(config["gpu_mem_fraction"])
        if "cpu_mem_fraction" in config:
            self.setCPUFract(config["cpu_mem_fraction"])
        if "use_phase_margin" in config:
            self.setPhasePaganinRequested(config["use_phase_margin"])
        if "postproc" in config:
            self.setPostProcConfiguration(config["postproc"])
        if "new_output_file_format" in config:
            new_output_file_format = config["new_output_file_format"]
            if new_output_file_format == "":
                new_output_file_format = None
            self.setNabuFileFormat(new_output_file_format)
        if "new_output_location" in config:
            new_output_location = config["new_output_location"]
            if new_output_location == "":
                new_output_location = None
            self.setNabuOutputLocation(new_output_location)
        output_dir_mode = config.get("output_dir_mode", None)
        if output_dir_mode is not None:
            self._outputLocationWidget.setOutputDirMode(output_dir_mode)
        overwrite_output_location = config.get("overwrite_output_location", None)
        if overwrite_output_location is not None:
            self.setRedefineOutputLocation(overwrite_output_location)

    def setPostProcConfiguration(self, config):
        if "output_histogram" in config:
            self.setHistogramRequested(requested=bool(config["output_histogram"]))

    def setConfigurationLevel(self, level):
        level = ConfigurationLevel.from_value(level)
        _NabuStageConfigBase.setConfigurationLevel(self, level)

    def getConfigurationLevel(self):
        return self._configuration_level


class NabuVolumeWindow(qt.QMainWindow):
    """
    Widget to define settings for a volume reconstruction
    """

    sigConfigChanged = qt.Signal()
    """Signal emitted when configuration change"""

    def __init__(self, parent):
        qt.QMainWindow.__init__(self, parent=parent)

        self.setWindowFlags(qt.Qt.Widget)

        self._mainWidget = NabuVolumeWidget(parent=self)
        self.setCentralWidget(self._mainWidget)

        # add toolbar
        toolbar = qt.QToolBar(self)
        toolbar.setMovable(False)
        toolbar.setFloatable(False)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        # add configuration mode
        self.__configurationModesAction = qt.QAction(self)
        self.__configurationModesAction.setCheckable(False)
        menu = qt.QMenu(self)
        self.__configurationModesAction.setMenu(menu)
        toolbar.addAction(self.__configurationModesAction)

        self.__configurationModesGroup = qt.QActionGroup(self)
        self.__configurationModesGroup.setExclusive(True)
        self.__configurationModesGroup.triggered.connect(self._userModeChanged)

        self._minimalisticAction = MinimalisticConfigurationAction(toolbar)
        menu.addAction(self._minimalisticAction)
        self.__configurationModesGroup.addAction(self._minimalisticAction)
        self._basicConfigAction = BasicConfigurationAction(toolbar)
        menu.addAction(self._basicConfigAction)
        self.__configurationModesGroup.addAction(self._basicConfigAction)
        self._expertConfiguration = ExpertConfigurationAction(toolbar)
        menu.addAction(self._expertConfiguration)
        self.__configurationModesGroup.addAction(self._expertConfiguration)

        # set up
        self._userModeChanged(action=self._basicConfigAction)
        self._basicConfigAction.setChecked(True)

        # connect signal / slot
        self._mainWidget.sigConfigChanged.connect(self._triggerSigConfigChanged)

    def _triggerSigConfigChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def _userModeChanged(self, action):
        self.__configurationModesAction.setIcon(action.icon())
        self.__configurationModesAction.setToolTip(action.tooltip())
        if action is self._basicConfigAction:
            level = ConfigurationLevel.OPTIONAL
        elif action is self._expertConfiguration:
            level = ConfigurationLevel.ADVANCED
        else:
            level = ConfigurationLevel.REQUIRED
        self._mainWidget.setConfigurationLevel(level)

    # expose API
    def getConfigurationLevel(self):
        self._mainWidget.getConfigurationLevel()

    def setConfigurationLevel(self, level):
        self._mainWidget.setConfigurationLevel(level=level)

    def getConfiguration(self):
        return self._mainWidget.getConfiguration()

    def setConfiguration(self, config):
        return self._mainWidget.setConfiguration(config=config)

    def setScan(self, scan):
        return self._mainWidget.setScan(scan=scan)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/03/2019"


import logging
import os
from collections import namedtuple
from typing import Union

import h5py
import numpy
from silx.io.url import DataUrl
from silx.utils.deprecation import deprecated
from silx.utils.enum import Enum as _Enum
from tomoscan.esrf.scan.utils import get_data

from tomwer.core.process.task import Task
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.spec import getParametersFromParOrInfo

from .anglemode import CorAngleMode
from .mode import AxisMode, AXIS_MODE_METADATAS
from .projectiontype import ProjectionType

from nabu.preproc.phase import PaganinPhaseRetrieval


_logger = logging.getLogger(__name__)


_calculation_conf = namedtuple(
    "_calculation_conf", ["projection_type", "paganin"]
)  # noqa

_WITH_PAG = "withpag"
_NO_PAG = "nopag"


DEFAULT_CMP_THETA = 5
DEFAULT_CMP_N_SUBSAMPLING_Y = 40
DEFAULT_CMP_NEAR_POS = 0
DEFAULT_CMP_NEAR_WIDTH = 40
DEFAULT_CMP_OVERSAMPLING = 4
DEFAULT_CMP_TAKE_LOG = True


class AxisCalculationInput(_Enum):
    """Define the different mode of input the user can have for axis calculation"""

    emission = _calculation_conf(ProjectionType.absorption, False)
    transmission = _calculation_conf(ProjectionType.transmission, False)
    transmission_pag = _calculation_conf(ProjectionType.transmission, True)

    def name(self):  # pylint: disable=E0102
        if self.value.paganin is True:
            return " ".join((self.value.projection_type.value, "paganin"))
        else:
            return self.value.projection_type.value

    def to_dict(self):
        pag_text = _WITH_PAG if self.value.paganin is True else _NO_PAG
        return "_".join((self.value.projection_type.value, pag_text))

    @classmethod
    def from_value(cls, value):
        if type(value) is str and len(value.split("_")) == 2:
            proj_type, pag = value.split("_")
            value_pag = True if pag == _WITH_PAG else False
            value_proj = ProjectionType.from_value(proj_type)
            return cls.from_value(_calculation_conf(value_proj, value_pag))
        elif isinstance(value, cls):
            return value
        else:
            for member in cls:
                if value in (member.value, member.name()):
                    return member
            raise ValueError(f"Cannot convert: {value}")


class AxisResource(object):
    """Helper for axis file relative stuff"""

    _PAGANIN_CONFIG = {"distance": 100e-3, "energy": 35, "delta_beta": 1e3}
    """Paganin configuration for axis calculation. To simplify we have only one
    static configuration for now. Otherwise complicate stuff"""

    def __init__(self, url):
        assert url is None or isinstance(url, DataUrl)
        assert url is None or url.is_valid()
        self.__url = url
        self.__raw_data = None
        self.__norme_data = None
        self.__norm_paganin = None

    def __str__(self):
        return f"{type(self)}, url: {self.__url.path() if self.__url else None}"

    @property
    def url(self):
        return self.__url

    @property
    def data(self):
        """

        :return: 2D numpy.array
        """
        if self.__url is None:
            return None
        if self.__raw_data is None:
            self.__raw_data = get_data(self.url)
        return self.__raw_data

    @data.setter
    def data(self, data):
        self.__raw_data = data

    @property
    def normalized_data(self):
        return self.__norme_data

    @normalized_data.setter
    def normalized_data(self, data):
        self.__norme_data = data

    def normalize_data(self, scan, log_):
        """Normalize data for axis calculation"""
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"scan is expected to be an instance of {TomwerScanBase}. {type(scan)} provided"
            )
        if self.__url is None:
            return None
        else:
            self.__norme_data = scan.data_flat_field_correction(data=self.data)
            if log_ is True:
                self.__norme_data[numpy.isnan(self.__norme_data)] = (
                    self.__norme_data.max()
                )
                if self.__norme_data.max() < 0:
                    _logger.error("max data value < 0 unable to compute log")
                else:
                    try:
                        data = self.__norme_data
                        data[data <= 1] = 1
                        self.__norme_data = -numpy.log(data)
                    except Exception as e:
                        _logger.error("Fail to apply log on radio" + e)
            else:
                self.__norme_data[numpy.isnan(self.__norme_data)] = (
                    self.__norme_data.min()
                )

        return self.__norme_data

    @property
    def normalized_data_paganin(self):
        """

        :return: data processed by the pagagnin phase retrieval
        :rtype: 2D numpy.array
        """
        if self.__url is None:
            return None
        if self.__norm_paganin is None:
            if self.normalized_data is None:
                raise ValueError("data should be normalized before applying " "paganin")
            else:
                data = self.normalized_data
                phase_retrieval = PaganinPhaseRetrieval(
                    data.shape, **AxisResource._PAGANIN_CONFIG
                )
                self.__norm_paganin = phase_retrieval.apply_filter(data)
        return self.__norm_paganin

    @normalized_data_paganin.setter
    def normalized_data_paganin(self, data):
        self.__norm_paganin = data

    def __eq__(self, other):
        if not isinstance(other, AxisResource):
            return False
        if self.url is None and other.url is None:
            return True
        elif self.url is None or other.url is None:
            return False
        else:
            return self.url == other.url


class AxisRP:
    """
    Configuration class for a tomwer :class:`AxisProcess`

    note: every modification on the parameters will process a call fo changed
    except `axis_url_1` and `axis_url_2` which will produce a call to the
    dedicated axis_url_changed
    """

    AXIS_POSITION_PAR_KEY = "ROTATION_AXIS_POSITION"
    """Key used for the axis position in par files"""

    _MANAGED_KEYS = (
        "MODE",
        "POSITION_VALUE",
        "CALC_INPUT_TYPE",
        "USE_SINOGRAM",
        "ANGLE_MODE",
        "SINOGRAM_LINE",
        "AXIS_URL_1",
        "AXIS_URL_2",
        "LOOK_AT_STDMAX",
        "NEAR_WX",
        "FINE_STEP_X",
        "SCALE_IMG2_TO_IMG1",
        "NEAR_POSITION",
        "SINOGRAM_SUBSAMPLING",
        "PADDING_MODE",
        "FLIP_LR",
        "COMPOSITE_OPTS",
        "SIDE",
        "COR_OPTIONS",
    )

    def __init__(self):
        self.__mode = AxisMode.manual
        """Mode used for defining the COR (center of rotation)"""
        self.__relative_value = None
        """Value of the center of rotation in [0; image_width].
        None is not processing"""
        self.__absolute_value = None
        """Value of the center of rotation in [-image_width/2; image_width/2].
        None is not processing"""
        self.__angle_mode = CorAngleMode.use_0_180
        """Angle to use for radios"""
        self.__estimated_cor = 0.0
        """Position to use for near calculation"""
        self.__axis_url_1 = AxisResource(url=None)
        """first data url to use for axis cor calculation"""
        self.__axis_url_2 = AxisResource(url=None)
        """second data url to use for axis cor calculation"""
        self.__calculation_input_type = AxisCalculationInput.transmission
        """Type of input (emission, absorption, with or without paganin)"""
        self.__use_sinogram = False
        """Do we want to use the sinogram of the radios for computing center
        of rotation"""
        self.__sinogram_line = "middle"
        """Line of the radios to use for getting the sinogram"""
        self.__sinogram_subsampling = 10
        """if use sinogram activate, we can use a subsampling to reduce
        computation time"""
        self.__look_at_stdmax = False
        """do the near search at X position which as the max Y column standard
        deviation"""
        self.__near_wx = 5
        """do the near search in an X window of size +-near_wx"""
        self.__fine_stepx = 0.1
        """shift step x for fine shifting image"""
        self.__scale_img2_to_img1 = False
        """do image scaling"""
        self.__padding_mode = None
        self.__frame_width = None
        self.__side = "right"
        """side of the cor. Requested by nabu cor algorithms growing-window
        and sliding-window"""
        self.__flip_lr = True
        self.__composite_options = {
            "theta": DEFAULT_CMP_THETA,
            "n_subsampling_y": DEFAULT_CMP_N_SUBSAMPLING_Y,
            "oversampling": DEFAULT_CMP_OVERSAMPLING,
            "take_log": DEFAULT_CMP_TAKE_LOG,
            "near_pos": DEFAULT_CMP_NEAR_POS,
            "near_width": DEFAULT_CMP_NEAR_WIDTH,
        }
        """specific options for composite cor search"""
        self.__extra_cor_options = ""
        """Automatic cor options as str. side, near_pos and near_width are provided independantly"""

    @property
    def mode(self):
        return self.__mode

    @mode.setter
    def mode(self, mode):
        if isinstance(mode, str):
            try:
                name = mode
                if name in ("global_", "global"):
                    name = AxisMode.global_.name
                elif name == "accurate":
                    _logger.info(
                        f"convert axis mode {name} to {AxisMode.centered.name} (renamed)"
                    )
                    name = AxisMode.centered.name
                elif name == "growing-window":
                    _logger.warning(
                        f"growing-window mode has been removed. Replace it by {AxisMode.growing_window_radios.value}"
                    )
                    name = AxisMode.growing_window_radios.name
                elif name == "sliding-window":
                    _logger.warning(
                        f"sliding-window mode has been removed. Replace it by {AxisMode.sliding_window_radios.value}"
                    )
                    name = AxisMode.sliding_window_radios.name
                try:
                    _mode = getattr(AxisMode, name)
                except Exception:
                    _mode = AxisMode.from_value(name)
            except Exception:
                raise ValueError(f"Fail to create axis mode from {mode}")
        else:
            if not isinstance(mode, AxisMode):
                raise TypeError(f"mode is expected to be an instance of {AxisMode}")
            _mode = mode
        self.__mode = _mode
        self.changed()

    @property
    def frame_width(self) -> Union[None, float]:
        return self.__frame_width

    @frame_width.setter
    def frame_width(self, value):
        if not isinstance(value, (type(None), float, int)):
            raise TypeError(
                f"Value is expected to be None or a float. Not {type(value)}"
            )
        if value is None:
            self.__frame_width = value
        else:
            self.__frame_width = float(value)

    @property
    def angle_mode(self):
        return self.__angle_mode

    @angle_mode.setter
    def angle_mode(self, mode):
        if type(mode) is str:
            mode = CorAngleMode.from_value(mode)

        if self.__angle_mode != mode:
            self.__angle_mode = mode
            self.changed()

    @property
    @deprecated(replacement="relative_cor_value", since_version="0.7")
    def value_ref_tomwer(self):
        return self.relative_cor_value

    @property
    def relative_cor_value(self):
        return self.__relative_value

    @property
    def absolute_cor_value(self):
        return self.__absolute_value

    def set_relative_value(self, value):
        if not isinstance(value, (int, float, str, type(None))):
            raise TypeError(
                f"value is expected to be an instance of {int} {float}, {str} or {None}. {type(value)} provided"
            )
        if value is None:
            changed = self.__relative_value is not None
        elif isinstance(value, str) and value == "...":
            changed = self.__relative_value != "..."
        else:
            changed = self.__relative_value != float(value)
        if changed:
            if value is None or (isinstance(value, str) and value == "..."):
                self.__relative_value = value
                self.__absolute_value = value
            else:
                self.__relative_value = float(value)
                if self.frame_width is not None:
                    self.__absolute_value = (
                        self.__relative_value + self.frame_width / 2.0
                    )
            self.changed()

    @deprecated(replacement="set_relative_value", since_version="0.7")
    def set_value_ref_tomwer(self, value):
        self.set_relative_value(value=value)

    @property
    @deprecated(replacement="absolute_cor_value", since_version="0.7")
    def value_ref_nabu(self):
        return self.__absolute_value

    @property
    def estimated_cor(self):
        return self.__estimated_cor

    @estimated_cor.setter
    def estimated_cor(self, value):
        if self.__estimated_cor != value:
            self.__estimated_cor = value
            self.changed()

    @property
    def side(self):
        return self.__side

    @side.setter
    def side(self, side):
        if side not in ("all", "left", "right", "center", "near"):
            raise ValueError(f"side '{side}' is not managed")
        if self.__side != side:
            self.__side = side
            self.changed()

    @property
    def axis_url_1(self):
        """the first file to be used for the axis calculation"""
        return self.__axis_url_1

    def _get_rsrc_frm_url(self, old_resource, new_url_):
        """
        Util function to compare new resource / url with an existing one.
        """
        new_resource = None
        changed = False
        if isinstance(new_url_, AxisResource):
            if old_resource != new_url_:
                changed = True
                new_resource = new_url_
        elif isinstance(new_url_, str):
            if new_url_ == "":
                new_resource = AxisResource(None)
            else:
                new_resource = AxisResource(DataUrl(path=new_url_))
            if old_resource != new_resource:
                changed = True
        else:
            assert isinstance(new_url_, DataUrl) or new_url_ is None
            new_resource = AxisResource(new_url_)
            if old_resource != new_resource:
                changed = True
        return new_resource, changed

    @axis_url_1.setter
    def axis_url_1(self, url_):
        new_resource, changed = self._get_rsrc_frm_url(
            old_resource=self.__axis_url_1, new_url_=url_
        )
        if changed is True:
            self.__axis_url_1 = new_resource
            self.axis_urls_changed()

    @property
    def axis_url_2(self):
        """the second file to be used for the axis calculation"""
        return self.__axis_url_2

    @axis_url_2.setter
    def axis_url_2(self, url_):
        new_resource, changed = self._get_rsrc_frm_url(
            old_resource=self.__axis_url_2, new_url_=url_
        )
        if changed is True:
            self.__axis_url_2 = new_resource
            self.axis_urls_changed()

    @property
    def flip_lr(self):
        return self.__flip_lr

    @flip_lr.setter
    def flip_lr(self, flip: bool):
        if not isinstance(flip, bool):
            raise TypeError("flip should be a boolean")
        self.__flip_lr = flip

    @property
    def projection_type(self):
        return self.__calculation_input_type.value.projection_type

    @property
    def paganin_preproc(self):
        return self.__calculation_input_type.value.paganin

    @property
    def calculation_input_type(self):
        return self.__calculation_input_type

    @calculation_input_type.setter
    def calculation_input_type(self, type_):
        assert isinstance(type_, AxisCalculationInput)
        value = AxisCalculationInput.from_value(type_)
        if value != self.__calculation_input_type:
            self.__calculation_input_type = type_
            self.changed()

    @property
    def use_sinogram(self):
        return self.__use_sinogram

    @use_sinogram.setter
    def use_sinogram(self, sinogram):
        if self.__use_sinogram != sinogram:
            self.__use_sinogram = sinogram
            self.changed()

    @property
    def sinogram_line(self):
        return self.__sinogram_line

    @sinogram_line.setter
    def sinogram_line(self, line):
        if line == "":
            if self.__sinogram_line is not None:
                self.__sinogram_line = None
                self.changed()
        else:
            if line == "middle":
                line = line
            elif line is not None:
                line = int(line)
            if self.__sinogram_line != line:
                self.__sinogram_line = line
                self.changed()

    @property
    def sinogram_subsampling(self) -> int:
        """Subsample radio to speed up processing of the sinogram generation"""
        return self.__sinogram_subsampling

    @sinogram_subsampling.setter
    def sinogram_subsampling(self, subsampling: int) -> None:
        subsampling = int(subsampling)
        if subsampling != self.__sinogram_subsampling:
            self.__sinogram_subsampling = subsampling
            self.changed()

    @property
    def look_at_stdmax(self):
        return self.__look_at_stdmax

    @look_at_stdmax.setter
    def look_at_stdmax(self, stdmax):
        self.__look_at_stdmax = stdmax

    @property
    def near_wx(self):
        return self.__near_wx

    @near_wx.setter
    def near_wx(self, width):
        if self.__near_wx != width:
            self.__near_wx = width
            self.changed()

    @property
    def fine_step_x(self):
        return self.__fine_stepx

    @fine_step_x.setter
    def fine_step_x(self, step_size):
        if self.__fine_stepx != step_size:
            self.__fine_stepx = step_size
            self.changed()

    @property
    def scale_img2_to_img1(self):
        return self.__scale_img2_to_img1

    @scale_img2_to_img1.setter
    def scale_img2_to_img1(self, scale):
        if self.__scale_img2_to_img1 != scale:
            self.__scale_img2_to_img1 = scale
            self.changed()

    @property
    def extra_cor_options(self) -> str:
        return self.__extra_cor_options

    @extra_cor_options.setter
    def extra_cor_options(self, cor_options: str):
        if not isinstance(cor_options, str):
            raise TypeError(f"{type(cor_options)} provided when {str} expected")
        else:
            clean_opts_str = (
                cor_options.replace(" ", "").strip(";").rstrip(";").replace(";", " ; ")
            )
            self.__extra_cor_options = clean_opts_str
            self.changed()

    @property
    def composite_options(self) -> dict:
        """return specific options for composite cor search"""
        return self.__composite_options

    @composite_options.setter
    def composite_options(self, opts: dict) -> None:
        """
        :param dict opts: options to use for the composite COR search
        :raises:
            * KeyError if some provided keys are not handled
            * TypeError if opts is not an instance of dictionary
        """
        if not isinstance(opts, dict):
            raise TypeError(
                f"opts is expected to be an instance of dict not {type(opts)}"
            )
        # insure backward compatibility
        if "subsampling_y" in opts:
            opts["n_subsampling_y"] = opts.pop("subsampling_y")
        for key in opts.keys():
            if key not in (
                "theta",
                "oversampling",
                "n_subsampling_y",
                "take_log",
                "near_pos",
                "near_width",
            ):
                raise KeyError(f"{key} is not recogized")
        self.__composite_options = opts

    @property
    def padding_mode(self):
        return self.__padding_mode

    @padding_mode.setter
    def padding_mode(self, mode):
        if self.__padding_mode != mode:
            self.__padding_mode = mode
            self.changed()

    def changed(self):
        """callback to overwrite when the paramter value changed"""
        pass

    def set_position_frm_par_file(self, file_path, entry=None, force=False):
        """Set position value from a file

        :param str file_path: path to the par file
        :param bool force: if True, force the mode to read and set
                           `position_value` to None if failed to read the file.
                           Otherwise only `mode` and `position_value` when the
                           file read succeed.
        """

        def get_position_in_file():
            if not os.path.isfile(file_path):
                raise ValueError(file_path, "is not a valid file path")

            if h5py.is_hdf5(file_path):
                return self.get_cor_position_frm_tomwer_file(
                    file_path=file_path, entry=entry
                )
            elif file_path.lower().endswith(".par") or file_path.lower().endswith(
                ".info"
            ):
                _dict = getParametersFromParOrInfo(file_path)
                if self.AXIS_POSITION_PAR_KEY.lower() not in _dict:
                    err = f"fail to find `{self.AXIS_POSITION_PAR_KEY}` in {file_path}"
                    _logger.info(err)
                    raise KeyError(err)
                else:
                    return _dict[self.AXIS_POSITION_PAR_KEY.lower()]
            else:
                with open(file_path, mode="r") as fd:
                    l1 = fd.readline()
                    values = l1.split(" ")
                    if len(values) > 0:
                        try:
                            res = float(values[0])
                        except Exception:
                            _logger.info(
                                f"fail to load center of rotation from {file_path}"
                            )
                            return None
                        else:
                            return res

        if force:
            try:
                position = get_position_in_file()
            except (KeyError, ValueError) as e:
                _logger.warning(e)
                position = None
        else:
            position = get_position_in_file()

        changed = self.mode != AxisMode.read or position != self.relative_cor_value
        self.mode = AxisMode.read
        if position is None:
            self.set_relative_value(None)
        else:
            self.set_relative_value(float(position))

        if changed:
            self.changed()

    @staticmethod
    def get_cor_position_frm_tomwer_file(
        file_path: str, entry=None
    ) -> Union[None, float]:
        """
        Parse the given .h5 file, looking for axis processes. If found some
        then return the first value discover

        :param str file_path: tomwer processes file (.h5)
        :return: center of rotation or none
        :rtype: Union[None, float]
        """
        from .axis import AxisTask  # avoid cyclic import

        processes = Task.get_processes_frm_type(
            process_file=file_path, process_type=AxisTask, entry=entry
        )
        if len(processes) == 0:
            _logger.warning("failed to find any information on center " "of rotation")
            return None
        elif len(processes) > 1:
            _logger.info("found more than one value computed for " "center of rotation")
        _logger.info(
            f"load center of rotation from {file_path}: {processes[0].results['center_of_rotation']}"
        )
        cor = processes[0].results["center_of_rotation"]
        if cor in [None, ""]:
            return None
        else:
            return float(cor)

    def n_url(self):
        """

        :return: number of available url from url_1, url_2
        """
        n_url = 0
        if self.axis_url_1 and self.axis_url_1.url:
            n_url += 1
        if self.axis_url_2 and self.axis_url_2.url:
            n_url += 1
        return n_url

    def to_dict(self):
        # keep octave compatibility
        axis_urls_1 = self.axis_url_1.url
        if axis_urls_1 is None:
            axis_urls_1 = ""
        else:
            axis_urls_1 = axis_urls_1.path()
        axis_urls_2 = self.axis_url_2.url
        if axis_urls_2 is None:
            axis_urls_2 = ""
        else:
            axis_urls_2 = axis_urls_2.path()

        _dict = {
            "MODE": self.mode.value,
            "POSITION_VALUE": self.relative_cor_value,
            "CALC_INPUT_TYPE": self.calculation_input_type.to_dict(),
            "ANGLE_MODE": self.angle_mode.value,
            "USE_SINOGRAM": self.use_sinogram,
            "SINOGRAM_LINE": self.sinogram_line if self.use_sinogram else "",
            "SINOGRAM_SUBSAMPLING": self.sinogram_subsampling,
            "AXIS_URL_1": axis_urls_1,
            "AXIS_URL_2": axis_urls_2,
            "LOOK_AT_STDMAX": self.look_at_stdmax,
            "NEAR_WX": self.near_wx,
            "FINE_STEP_X": self.fine_step_x,
            "SCALE_IMG2_TO_IMG1": self.scale_img2_to_img1,
            "NEAR_POSITION": self.estimated_cor,
            "PADDING_MODE": self.padding_mode,
            "FLIP_LR": self.flip_lr,
            "COMPOSITE_OPTS": self.composite_options,
            "SIDE": self.side,
            "COR_OPTIONS": self.extra_cor_options,
        }
        return _dict

    @staticmethod
    def from_dict(_dict):
        axis = AxisRP()
        axis.load_from_dict(_dict=_dict)
        return axis

    def load_from_dict(self, _dict):
        # Convert managed keys to upper case
        _dict = {
            key.upper() if key.upper() in self._MANAGED_KEYS else key: value
            for key, value in _dict.items()
        }

        if "MODE" in _dict:
            self.mode = _dict["MODE"]
        if "POSITION_VALUE" in _dict:
            self.set_relative_value(_dict["POSITION_VALUE"])
        if "CALC_INPUT_TYPE" in _dict:
            self.calculation_input_type = AxisCalculationInput.from_value(
                _dict["CALC_INPUT_TYPE"]
            )
        if "USE_SINOGRAM" in _dict:
            self.use_sinogram = _dict["USE_SINOGRAM"]
        if "ANGLE_MODE" in _dict:
            self.angle_mode = CorAngleMode.from_value(_dict["ANGLE_MODE"])
        if "SINOGRAM_LINE" in _dict:
            self.sinogram_line = _dict["SINOGRAM_LINE"]
        if "AXIS_URL_1" in _dict:
            self.axis_url_1 = _dict["AXIS_URL_1"]
        if "AXIS_URL_2" in _dict:
            self.axis_url_2 = _dict["AXIS_URL_2"]
        if "LOOK_AT_STDMAX" in _dict:
            self.look_at_stdmax = _dict["LOOK_AT_STDMAX"]
        if "NEAR_WX" in _dict:
            self.near_wx = _dict["NEAR_WX"]
        if "FINE_STEP_X" in _dict:
            self.fine_step_x = _dict["FINE_STEP_X"]
        if "SCALE_IMG2_TO_IMG1" in _dict:
            self.scale_img2_to_img1 = _dict["SCALE_IMG2_TO_IMG1"]
        if "NEAR_POSITION" in _dict:
            self.estimated_cor = float(_dict["NEAR_POSITION"])
        if "SINOGRAM_SUBSAMPLING" in _dict:
            self.sinogram_subsampling = _dict["SINOGRAM_SUBSAMPLING"]
        if "PADDING_MODE" in _dict:
            self.padding_mode = _dict["PADDING_MODE"]
        if "FLIP_LR" in _dict:
            self.flip_lr = bool(_dict["FLIP_LR"])
        if "COMPOSITE_OPTS" in _dict:
            self.composite_options = _dict["COMPOSITE_OPTS"]
        if "SIDE" in _dict:
            self.side = _dict["SIDE"]
        self.extra_cor_options = _dict.get("COR_OPTIONS", "")

    def copy(self, axis_params, copy_axis_url=True, copy_flip_lr=True):
        assert isinstance(axis_params, AxisRP)
        self.mode = axis_params.mode
        self.frame_width = axis_params.frame_width
        self.set_relative_value(axis_params.relative_cor_value)
        self.calculation_input_type = axis_params.calculation_input_type
        self.use_sinogram = axis_params.use_sinogram
        self.angle_mode = axis_params.angle_mode
        self.sinogram_line = axis_params.sinogram_line
        self.sinogram_subsampling = axis_params.sinogram_subsampling
        self.look_at_stdmax = axis_params.look_at_stdmax
        self.near_wx = axis_params.near_wx
        self.fine_step_x = axis_params.fine_step_x
        self.scale_img2_to_img1 = axis_params.scale_img2_to_img1
        self.estimated_cor = axis_params.estimated_cor
        self.padding_mode = axis_params.padding_mode
        self.side = axis_params.side
        self.composite_options = axis_params.composite_options
        self.extra_cor_options = axis_params.extra_cor_options
        if copy_axis_url:
            self.axis_url_1 = axis_params.axis_url_1
            self.axis_url_2 = axis_params.axis_url_2
        if copy_flip_lr:
            self.flip_lr = axis_params.flip_lr

    def __str__(self):
        return str(self.to_dict())

    def axis_urls_changed(self):
        """Callback when the axis url change"""
        pass

    def get_simple_str(self):
        """
        special information as a str for mode able to handle both sinogram and radios
        """
        results = f"{self.mode.value}"
        if self.mode in (
            AxisMode.growing_window_radios,
            AxisMode.growing_window_sinogram,
            AxisMode.sliding_window_radios,
            AxisMode.sliding_window_sinogram,
        ):
            extra_info = f"side: {self.side}, use sinogram: {self.use_sinogram}"
            results = ", ".join((results, extra_info))
        return results

    def get_nabu_cor_options_as_dict(self) -> str:
        options = {}
        if self.mode is AxisMode.near:
            self.side = "near"
        request_side = len(AXIS_MODE_METADATAS[self.mode].valid_sides) > 0
        if request_side:

            if self.side == "near":
                options["side"] = self.composite_options.get(
                    "near_pos", self.estimated_cor
                )
            else:
                options["side"] = self.side

        if self.side == "near":
            near_width = self.composite_options.get("near_width", 20.0)
            options["near_width"] = near_width

        # append "extra_cor_options" to already handled cor options
        extra_cor_options = self.extra_cor_options.replace(" ", "")

        if extra_cor_options != "":
            for opt in self.extra_cor_options.replace(" ", "").split(";"):
                if len(opt.split("=")) == 2:
                    key, value = opt.split("=")
                    options[key] = value
                else:
                    _logger.info(f"ignore option {opt}. Invalid synthax")
        return options

    @deprecated(replacement="get_nabu_cor_options_as_str", since_version="1.1")
    def get_nabu_cor_options(self) -> str:
        return self.get_nabu_cor_options_as_str()

    def get_nabu_cor_options_as_str(self) -> str:
        """return cor option for nabu"""

        def cast_key_value(key, value):
            if key in ("side",):
                return f"{key}='{value}'"
            else:
                return f"{key}={value}"

        return " ; ".join(
            [
                cast_key_value(key, value)
                for key, value in self.get_nabu_cor_options_as_dict().items()
            ]
        )



import gc
import os
import pickle
import tempfile

import numpy
import pytest
from nxtomomill.nexus.nxtomo import NXtomo
from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import SignalListener, TestCaseQt
from nxtomo.nxobject.nxdetector import ImageKey

from orangecontrib.tomwer.widgets.edit.NXtomoEditorOW import NXtomoEditorOW
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestNXtomoEditorOW(TestCaseQt):
    """Test that the NXtomoEditorOW widget work correctly. Processing test are done in the core module. gui test are done in the tomwer.gui.edit module"""

    def setUp(self):
        super().setUp()
        self._window = NXtomoEditorOW()

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        self._window = None
        gc.collect()

    def test(self):
        self._window.show()
        self.qWaitForWindowExposed(self._window)

    def getDefaultConfig(self) -> dict:
        return {
            "instrument.beam.energy": (5.9, False),
            "instrument.detector.distance": (2.4, True),
            "instrument.detector.field_of_view": ("Full", False),
            "instrument.detector.x_pixel_size": (0.023, True),
            "instrument.detector.y_pixel_size": (0.025, True),
            "instrument.detector.x_flipped": (True, True),
            "instrument.detector.y_flipped": (False, False),
        }

    def testSerialization(self):
        self._window.setConfiguration(self.getDefaultConfig())
        assert self._window.getConfiguration() == self.getDefaultConfig()
        pickle.dumps(self._window.getConfiguration())

    def testLiteralDumps(self):
        self._window.setConfiguration(self.getDefaultConfig())
        assert self._window.getConfiguration() == self.getDefaultConfig()
        literal_dumps(self._window.getConfiguration())

    def testAutomation(self):
        """
        test that if some field are locked then this will overwrite the Nxtomo and send the signal
        """
        signal_listener = SignalListener()
        self._window.sigScanReady.connect(signal_listener)
        # set up the widget to define and lock distance, energy and x pixel size
        distance_widget = self._window.widget.mainWidget._distanceMetricEntry
        distance_widget.setValue(0.6)
        distance_widget.setUnit("mm")
        distance_locker = self._window.widget.mainWidget._distanceLB
        distance_locker.setLock(True)
        energy_widget = self._window.widget.mainWidget._energyEntry
        energy_widget.setValue(88.058)
        energy_locker = self._window.widget.mainWidget._energyLockerLB
        energy_locker.setLock(True)
        x_pixel_widget = self._window.widget.mainWidget._xPixelSizeMetricEntry
        x_pixel_widget.setValue(45)
        x_pixel_widget.setUnit("nm")
        x_pixel_locker = self._window.widget.mainWidget._xPixelSizeLB
        x_pixel_locker.setLock(True)

        with tempfile.TemporaryDirectory() as tmp_dir:
            # 1.0 create nx tomos with raw data
            nx_tomo = NXtomo()
            nx_tomo.instrument.detector.x_pixel_size = 0.023
            nx_tomo.instrument.detector.y_pixel_size = 0.025
            nx_tomo.instrument.detector.field_of_view = "full"
            nx_tomo.instrument.detector.distance = 2.4
            nx_tomo.instrument.detector.x_flipped = False
            nx_tomo.instrument.detector.y_flipped = True
            nx_tomo.energy = 5.9
            nx_tomo.instrument.detector.image_key_control = [
                ImageKey.PROJECTION.value
            ] * 12
            nx_tomo.instrument.detector.data = numpy.empty(shape=(12, 10, 10))
            nx_tomo.sample.rotation_angle = numpy.linspace(0, 20, num=12)

            file_path = os.path.join(tmp_dir, "nxtomo.nx")
            entry = "entry0000"
            nx_tomo.save(
                file_path=file_path,
                data_path=entry,
            )
            # 2.0 set scan to the nxtomo-editor
            scan = NXtomoScan(file_path, entry)
            self._window._setScan(scan=scan)
            self.qapp.processEvents()
            # 3.0 check results are as expected
            # make sure the scan has been re-emitted
            assert signal_listener.callCount() == 1
            # make sure the edition of the parameters have been done and only those
            overwrite_nx_tomo = NXtomo().load(
                file_path=file_path,
                data_path=entry,
                detector_data_as="as_numpy_array",
            )
            numpy.testing.assert_almost_equal(
                overwrite_nx_tomo.instrument.detector.x_pixel_size.value, 45e-9
            )
            assert (
                overwrite_nx_tomo.instrument.detector.y_pixel_size.value
                == nx_tomo.instrument.detector.y_pixel_size.value
            )
            assert (
                overwrite_nx_tomo.instrument.detector.field_of_view
                == nx_tomo.instrument.detector.field_of_view
            )
            numpy.testing.assert_almost_equal(
                overwrite_nx_tomo.instrument.detector.distance.value, 6.0e-4
            )
            assert (
                overwrite_nx_tomo.instrument.detector.x_flipped
                == nx_tomo.instrument.detector.x_flipped
            )
            assert (
                overwrite_nx_tomo.instrument.detector.y_flipped
                == nx_tomo.instrument.detector.y_flipped
            )
            numpy.testing.assert_almost_equal(overwrite_nx_tomo.energy.value, 88.058)
            numpy.testing.assert_array_almost_equal(
                overwrite_nx_tomo.instrument.detector.data,
                nx_tomo.instrument.detector.data,
            )



import os

from tomwer.core.tomwer_object import TomwerObject


class TomwerVolumeBase(TomwerObject):
    def _clear_heavy_cache(self):
        """util user"""
        self.data = None
        self.metadata = None

    @staticmethod
    def format_output_location(location: str, volume):
        if not isinstance(volume, TomwerVolumeBase):
            raise TypeError(
                f"volume is expected to be an instance of {TomwerVolumeBase}"
            )

        keywords = {
            "volume_data_parent_folder": volume.volume_data_parent_folder(),
        }

        # filter necessary keywords
        def get_necessary_keywords():
            import string

            formatter = string.Formatter()
            return [field for _, field, _, _ in formatter.parse(location) if field]

        requested_keywords = get_necessary_keywords()

        def keyword_needed(pair):
            keyword, _ = pair
            return keyword in requested_keywords

        keywords = dict(filter(keyword_needed, keywords.items()))
        location = os.path.abspath(location.format(**keywords))
        return location

    def volume_data_parent_folder(self):
        if self.data_url is None:  # pylint: disable=E1101
            raise ValueError("data_url doesn't exists")
        else:
            return os.path.dirname(self.data_url.file_path())  # pylint: disable=E1101

    def __str__(self) -> str:
        try:
            return self.get_identifier().to_str()
        except Exception:
            return super().__str__()



from processview.core.dataset import Dataset


class TomwerObject(Dataset):
    """Common tomwer object"""

    _DICT_USER_STITCHING_METADATA = "user_def_stitching_metadata"

    def __init__(self) -> None:
        super().__init__()
        self._cast_volume = None
        self._stitching_metadata = None

    def _clear_heavy_cache(self):
        """util function to clear some heavy object from the cache"""
        raise NotImplementedError()

    def clear_caches(self):
        pass

    @property
    def cast_volume(self):
        # for now this is used as an east way to cache the identifier and provide it to the remaining of the orange canvas.
        # but this is a wrong designa and should be removed at one point.
        return self._cast_volume

    @cast_volume.setter
    def cast_volume(self, volume):
        from tomwer.core.volume.volumebase import TomwerVolumeBase

        if not (volume is None or isinstance(volume, TomwerVolumeBase)):
            from tomwer.core.volume.volumefactory import (
                VolumeFactory,
            )  # avoid cyclic import

            volume = VolumeFactory.create_tomo_object_from_identifier(identifier=volume)
        self._cast_volume = volume

    @property
    def stitching_metadata(self):
        return self._stitching_metadata

    @stitching_metadata.setter
    def stitching_metadata(self, metadata):
        from tomwer.core.process.stitching.metadataholder import StitchingMetadata

        if metadata is not None and not isinstance(metadata, StitchingMetadata):
            raise TypeError(
                f"metadata is expected to be an optional instance of {StitchingMetadata}. Not {type(metadata)}"
            )
        self._stitching_metadata = metadata



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "19/07/2018"

import functools
import logging

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output
from processview.core.manager import DatasetState, ProcessManager
from silx.gui import qt

import tomwer.core.process.control.timer
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.utils import docstring

_logger = logging.getLogger(__name__)


class _TimerWidget(qt.QWidget):
    def __init__(self, parent, _time=None):
        if _time is not None:
            assert type(_time) is int
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())

        self.layout().addWidget(qt.QLabel("time to wait (in sec):", parent=self), 0, 0)
        self._timeLE = qt.QSpinBox(parent=self)
        self._timeLE.setMinimum(0)
        self._timeLE.setValue(_time or 1)
        self.layout().addWidget(self._timeLE, 0, 1)

        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 1, 0)

        # expose API
        self.timeChanged = self._timeLE.valueChanged


class TimerOW(SuperviseOW):
    name = "timer"
    id = "orange.widgets.tomwer.filterow"
    description = (
        "Simple widget which wait for a defined amont of time and" "release the data"
    )
    icon = "icons/time.png"
    priority = 200
    keywords = ["control", "timer", "wait", "data"]

    want_main_area = True
    resizing_enabled = True
    _waiting_time = Setting(int(1))

    ewokstaskclass = tomwer.core.process.control.timer.TimerTask

    class Inputs:
        data = Input(name="data", type=TomwerScanBase, multiple=True)

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        """ """
        SuperviseOW.__init__(self, parent)
        self._widget = _TimerWidget(parent=self, _time=self._waiting_time)
        self._widget.setContentsMargins(0, 0, 0, 0)

        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self._widget)

        self._widget.timeChanged.connect(self._updateTime)

    @Inputs.data
    def process(self, scan, *args, **kwargs):
        if scan is None:
            return

        ProcessManager().notify_dataset_state(
            dataset=scan,
            process=self,
            state=DatasetState.ON_GOING,
        )

        callback = functools.partial(self._get_out, scan)
        _timer = qt.QTimer(self)
        _timer.singleShot(self._waiting_time * 1000, callback)

    def _updateTime(self, newTime):
        self._waiting_time = newTime

    def _get_out(self, scan):
        ProcessManager().notify_dataset_state(
            dataset=scan,
            process=self,
            state=DatasetState.SUCCEED,
        )
        self.Outputs.data.send(scan)

    @docstring(SuperviseOW)
    def reprocess(self, dataset):
        self.process(dataset)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "04/02/2020"


from typing import Optional

from silx.gui import qt

from tomwer.core.scan.scanbase import TomwerScanBase


class ScanNameLabelAndShape(qt.QWidget):
    """Scan to display the scan name"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent=parent)
        self.setLayout(qt.QHBoxLayout())
        self.setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        label = qt.QLabel("scan: ", self)
        label.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self.layout().addWidget(label)
        self._scanNameLabel = qt.QLabel("", self)
        self._scanNameLabel.setAlignment(qt.Qt.AlignLeft)
        self._scanNameLabel.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )
        self.layout().addWidget(self._scanNameLabel)

        self._shapeLabel = qt.QLabel("", self)
        self._shapeLabel.setAlignment(qt.Qt.AlignLeft)
        self._shapeLabel.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(self._shapeLabel)

        # set up
        self.clear()

    def setScan(self, scan: Optional[TomwerScanBase]):
        if scan is None or scan.path is None:
            self.clear()
        elif not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"Scan is expected to be an  instance of {TomwerScanBase}. Get {type(scan)} instead"
            )
        else:
            assert isinstance(scan, TomwerScanBase)
            self._scanNameLabel.setText(scan.get_identifier().short_description())
            self._scanNameLabel.setToolTip(scan.get_identifier().to_str())

            shape_x = scan.dim_1 if scan.dim_1 is not None else "?"
            shape_y = scan.dim_2 if scan.dim_2 is not None else "?"
            self._shapeLabel.setText(f"dims: x={shape_x}, y={shape_y}")

    def clear(self):
        self._scanNameLabel.setText("-")
        self._scanNameLabel.setToolTip("")
        self._shapeLabel.setText("")



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "19/10/2021"


import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.settings import SlurmSettings
from tomwer.gui.cluster.slurm import (
    SlurmSettingsMode,
    SlurmSettingsWidget,
    SlurmSettingsWindow,
)
from tomwer.tests.utils import skip_gui_test
from tomwer.tests.conftest import qtapp  # noqa F401


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestSlurmWidget(TestCaseQt):
    def setUp(self):
        TestCaseQt.setUp(self)
        self.slurmWidget = SlurmSettingsWidget(parent=None)

    def tearDown(self):
        self.slurmWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.slurmWidget.close()
        self.slurmWidget = None

    def testGetConfiguration(self):
        dict_res = self.slurmWidget.getConfiguration()
        expected_dict = {
            "cpu-per-task": SlurmSettings.N_CORES_PER_TASK,
            "n_tasks": SlurmSettings.N_TASKS,
            "memory": SlurmSettings.MEMORY_PER_WORKER,
            "partition": "",
            "n_gpus": SlurmSettings.N_GPUS_PER_WORKER,
            "job_name": "tomwer_{scan}_-_{process}_-_{info}",
            "n_jobs": 1,
            "modules": ("tomotools/stable",),
            "sbatch_extra_params": {
                "export": "NONE",
                "gpu_card": None,
            },
        }
        assert dict_res == expected_dict, f"{dict_res} vs {expected_dict}"

    def testSetConfiguration(self):
        self.slurmWidget.setConfiguration(
            {
                "cpu-per-task": 2,
                "n_tasks": 3,
                "memory": 156,
                "partition": "test-queue",
                "n_gpus": 5,
                "modules": "mymodule, mysecond/10.3",
                "sbatch_extra_params": {
                    "export": "ALL",
                },
            }
        )

        assert self.slurmWidget.getNCores() == 2
        assert self.slurmWidget.getNWorkers() == 3
        assert self.slurmWidget.getMemory() == 156
        assert self.slurmWidget.getQueue() == "test-queue"
        assert self.slurmWidget.getNGPU() == 5
        assert self.slurmWidget.getSBatchExtraParams() == {
            "export": "ALL",
            "gpu_card": None,
        }


def test_SlurmSettingsWindow(qtapp):  # noqa F811
    """
    test that slurm SettingsWindow can load all the different slurm settings modes
    """
    widget = SlurmSettingsWindow()
    for mode in SlurmSettingsMode:
        widget.setCurrentSettingsMode(mode)
        widget.getConfiguration()
